# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import List, Union

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer
from torch import Tensor

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import coil_combination_method
from atommic.collections.quantitative.nn.base import BaseqMRIReconstructionModel, SignalForwardModel
from atommic.collections.quantitative.nn.qvarnet_base.qvarnet_block import qVarNetBlock
from atommic.collections.quantitative.parts.transforms import R2star_B0_S0_phi_mapping
from atommic.collections.reconstruction.nn.unet_base.unet_block import NormUnet
from atommic.collections.reconstruction.nn.varnet_base.varnet_block import VarNetBlock
from atommic.core.classes.common import typecheck

__all__ = ["qVarNet"]


class qVarNet(BaseqMRIReconstructionModel):
    """Implementation of the quantitative End-to-end Variational Network (qVN), as presented in [Zhang2022]_.

    References
    ----------
    .. [Zhang2022] Zhang C, Karkalousos D, Bazin PL, Coolen BF, Vrenken H, Sonke JJ, Forstmann BU, Poot DH, Caan MW.
        A unified model for reconstruction and R2* mapping of accelerated 7T data using the quantitative recurrent
        inference machine. NeuroImage. 2022 Dec 1;264:119680.
    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        # init superclass
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)
        quantitative_module_dimensionality = cfg_dict.get("quantitative_module_dimensionality")
        if quantitative_module_dimensionality != 2:
            raise ValueError(
                f"Only 2D is currently supported for qMRI models.Found {quantitative_module_dimensionality}"
            )

        self.reconstruction_module = torch.nn.ModuleList([])

        self.use_reconstruction_module = cfg_dict.get("use_reconstruction_module")
        if self.use_reconstruction_module:
            self.reconstruction_module_num_cascades = cfg_dict.get("reconstruction_module_num_cascades")
            self.reconstruction_module_no_dc = cfg_dict.get("reconstruction_module_no_dc")

            for _ in range(self.reconstruction_module_num_cascades):
                self.reconstruction_module.append(
                    VarNetBlock(
                        NormUnet(
                            chans=cfg_dict.get("reconstruction_module_channels"),
                            num_pools=cfg_dict.get("reconstruction_module_pooling_layers"),
                            in_chans=cfg_dict.get("reconstruction_module_in_channels"),
                            out_chans=cfg_dict.get("reconstruction_module_out_channels"),
                            padding_size=cfg_dict.get("reconstruction_module_padding_size"),
                            normalize=cfg_dict.get("reconstruction_module_normalize"),
                        ),
                        fft_centered=self.fft_centered,
                        fft_normalization=self.fft_normalization,
                        spatial_dims=self.spatial_dims,
                        coil_dim=self.coil_dim - 1,
                        no_dc=self.reconstruction_module_no_dc,
                    )
                )

            self.dc_weight = torch.nn.Parameter(torch.ones(1))
            self.reconstruction_module_accumulate_predictions = cfg_dict.get(
                "reconstruction_module_accumulate_predictions"
            )

        quantitative_module_num_cascades = cfg_dict.get("quantitative_module_num_cascades")
        self.quantitative_module = torch.nn.ModuleList(
            [
                qVarNetBlock(
                    NormUnet(
                        chans=cfg_dict.get("quantitative_module_channels"),
                        num_pools=cfg_dict.get("quantitative_module_pooling_layers"),
                        in_chans=cfg_dict.get("quantitative_module_in_channels"),
                        out_chans=cfg_dict.get("quantitative_module_out_channels"),
                        padding_size=cfg_dict.get("quantitative_module_padding_size"),
                        normalize=cfg_dict.get("quantitative_module_normalize"),
                    ),
                    fft_centered=self.fft_centered,
                    fft_normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                    coil_dim=self.coil_dim,
                    no_dc=cfg_dict.get("quantitative_module_no_dc"),
                    linear_forward_model=SignalForwardModel(
                        sequence=cfg_dict.get("quantitative_module_signal_forward_model_sequence")
                    ),
                )
                for _ in range(quantitative_module_num_cascades)
            ]
        )

        self.quantitative_maps_regularization_factors = cfg_dict.get(
            "quantitative_maps_regularization_factors", [150.0, 150.0, 1000.0, 150.0]
        )

        self.accumulate_predictions = cfg_dict.get("quantitative_module_accumulate_predictions")

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        R2star_map_init: torch.Tensor,
        S0_map_init: torch.Tensor,
        B0_map_init: torch.Tensor,
        phi_map_init: torch.Tensor,
        TEs: List,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        initial_prediction: torch.Tensor,
        anatomy_mask: torch.Tensor,
        sampling_mask: torch.Tensor,
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> Union[List[List[Tensor]], List[Tensor]]:
        """
        Forward pass of the network.

        Parameters
        ----------
        R2star_map_init : torch.Tensor
            Initial R2* map of shape [batch_size, n_x, n_y].
        S0_map_init : torch.Tensor
            Initial S0 map of shape [batch_size, n_x, n_y].
        B0_map_init : torch.Tensor
            Initial B0 map of shape [batch_size, n_x, n_y].
        phi_map_init : torch.Tensor
            Initial phase map of shape [batch_size, n_x, n_y].
        TEs : List
            List of echo times.
        y : torch.Tensor
            Subsampled k-space data of shape [batch_size, n_echoes, n_coils, n_x, n_y, 2].
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2].
        initial_prediction : torch.Tensor
            Initial prediction of shape [batch_size, n_x, n_y, 2].
        anatomy_mask : torch.Tensor
            Brain mask of shape [batch_size, 1, n_x, n_y, 1].
        sampling_mask : torch.Tensor
            Sampling mask of shape [batch_size, 1, n_x, n_y, 1].
        sigma : float
            Standard deviation of the noise. Default is ``1.0``.

        Returns
        -------
        List of list of torch.Tensor or torch.Tensor
             If self.accumulate_loss is True, returns a list of all intermediate predictions.
             If False, returns the final estimate.
        """
        if self.use_reconstruction_module:
            cascades_echoes_predictions = []
            for echo in range(y.shape[1]):
                prediction = y[:, echo, ...].clone()
                for cascade in self.reconstruction_module:
                    # Forward pass through the cascades
                    prediction = cascade(prediction, y[:, echo, ...], sensitivity_maps, sampling_mask.squeeze(1))
                reconstruction_prediction = ifft2(
                    prediction,
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                )
                reconstruction_prediction = coil_combination_method(
                    reconstruction_prediction,
                    sensitivity_maps,
                    method=self.coil_combination_method,
                    dim=self.coil_dim - 1,
                )
                cascades_echoes_predictions.append(torch.view_as_complex(reconstruction_prediction))

            reconstruction_prediction = torch.stack(cascades_echoes_predictions, dim=1)
            if reconstruction_prediction.shape[-1] != 2:
                reconstruction_prediction = torch.view_as_real(reconstruction_prediction)

            y = fft2(
                coil_combination_method(
                    reconstruction_prediction.unsqueeze(self.coil_dim),
                    sensitivity_maps.unsqueeze(self.coil_dim - 1),
                    method=self.coil_combination_method,
                    dim=self.coil_dim - 1,
                ),
                self.fft_centered,
                self.fft_normalization,
                self.spatial_dims,
            )

            R2star_maps_init = []
            S0_maps_init = []
            B0_maps_init = []
            phi_maps_init = []
            for batch_idx in range(reconstruction_prediction.shape[0]):
                R2star_map_init, S0_map_init, B0_map_init, phi_map_init = R2star_B0_S0_phi_mapping(
                    reconstruction_prediction[batch_idx],
                    TEs,
                    anatomy_mask,
                    fft_centered=self.fft_centered,
                    fft_normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                )
                R2star_maps_init.append(R2star_map_init.squeeze(0))
                S0_maps_init.append(S0_map_init.squeeze(0))
                B0_maps_init.append(B0_map_init.squeeze(0))
                phi_maps_init.append(phi_map_init.squeeze(0))
            R2star_map_init = torch.stack(R2star_maps_init, dim=0).to(y)
            S0_map_init = torch.stack(S0_maps_init, dim=0).to(y)
            B0_map_init = torch.stack(B0_maps_init, dim=0).to(y)
            phi_map_init = torch.stack(phi_maps_init, dim=0).to(y)
        else:
            reconstruction_prediction = initial_prediction.clone()

        R2star_map_pred = R2star_map_init / self.quantitative_maps_regularization_factors[0]
        S0_map_pred = S0_map_init / self.quantitative_maps_regularization_factors[1]
        B0_map_pred = B0_map_init / self.quantitative_maps_regularization_factors[2]
        phi_map_pred = phi_map_init / self.quantitative_maps_regularization_factors[3]

        prediction = torch.stack([R2star_map_pred, S0_map_pred, B0_map_pred, phi_map_pred], dim=1)
        for cascade in self.quantitative_module:
            # Forward pass through the cascades
            prediction = cascade(
                prediction,
                y,
                sensitivity_maps,
                sampling_mask,
                TEs,
            )

        R2star_map_pred, S0_map_pred, B0_map_pred, phi_map_pred = (
            prediction[:, 0, ...] * self.quantitative_maps_regularization_factors[0],
            prediction[:, 1, ...] * self.quantitative_maps_regularization_factors[1],
            prediction[:, 2, ...] * self.quantitative_maps_regularization_factors[2],
            prediction[:, 3, ...] * self.quantitative_maps_regularization_factors[3],
        )

        return [
            reconstruction_prediction if self.use_reconstruction_module else torch.empty([]),
            R2star_map_pred,
            S0_map_pred,
            B0_map_pred,
            phi_map_pred,
        ]



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import math
from typing import Dict, List, Union

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer
from torch import Tensor

from atommic.collections.common.parts.fft import fft2
from atommic.collections.common.parts.utils import coil_combination_method, rnn_weights_init
from atommic.collections.quantitative.nn.base import BaseqMRIReconstructionModel, SignalForwardModel
from atommic.collections.quantitative.nn.qrim_base.qrim_block import qRIMBlock
from atommic.collections.quantitative.parts.transforms import R2star_B0_S0_phi_mapping
from atommic.collections.reconstruction.nn.rim_base.rim_block import RIMBlock
from atommic.core.classes.common import typecheck

__all__ = ["qCIRIM"]


class qCIRIM(BaseqMRIReconstructionModel):
    """Implementation of the quantitative Recurrent Inference Machines (qRIM), as presented in [Zhang2022]_.

    Also implements the qCIRIM model, which is a qRIM model with cascades.

    References
    ----------
    .. [Zhang2022] Zhang C, Karkalousos D, Bazin PL, Coolen BF, Vrenken H, Sonke JJ, Forstmann BU, Poot DH, Caan MW.
        A unified model for reconstruction and R2* mapping of accelerated 7T data using the quantitative recurrent
        inference machine. NeuroImage. 2022 Dec 1;264:119680.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`qCIRIM`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            Trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)
        quantitative_module_dimensionality = cfg_dict.get("quantitative_module_dimensionality")
        if quantitative_module_dimensionality != 2:
            raise ValueError(
                f"Only 2D is currently supported for qMRI models.Found {quantitative_module_dimensionality}"
            )

        quantitative_module_no_dc = cfg_dict.get("quantitative_module_no_dc")
        if not quantitative_module_no_dc:
            raise ValueError("qCIRIM does not support explicit DC component.")

        self.reconstruction_module = torch.nn.ModuleList([])

        self.use_reconstruction_module = cfg_dict.get("use_reconstruction_module")
        if self.use_reconstruction_module:
            self.reconstruction_module_recurrent_filters = cfg_dict.get("reconstruction_module_recurrent_filters")
            self.reconstruction_module_time_steps = 8 * math.ceil(cfg_dict.get("reconstruction_module_time_steps") / 8)
            self.reconstruction_module_no_dc = cfg_dict.get("reconstruction_module_no_dc")
            self.reconstruction_module_num_cascades = cfg_dict.get("reconstruction_module_num_cascades")

            for _ in range(self.reconstruction_module_num_cascades):
                self.reconstruction_module.append(
                    RIMBlock(
                        recurrent_layer=cfg_dict.get("reconstruction_module_recurrent_layer"),
                        conv_filters=cfg_dict.get("reconstruction_module_conv_filters"),
                        conv_kernels=cfg_dict.get("reconstruction_module_conv_kernels"),
                        conv_dilations=cfg_dict.get("reconstruction_module_conv_dilations"),
                        conv_bias=cfg_dict.get("reconstruction_module_conv_bias"),
                        recurrent_filters=self.reconstruction_module_recurrent_filters,
                        recurrent_kernels=cfg_dict.get("reconstruction_module_recurrent_kernels"),
                        recurrent_dilations=cfg_dict.get("reconstruction_module_recurrent_dilations"),
                        recurrent_bias=cfg_dict.get("reconstruction_module_recurrent_bias"),
                        depth=cfg_dict.get("reconstruction_module_depth"),
                        time_steps=self.reconstruction_module_time_steps,
                        conv_dim=cfg_dict.get("reconstruction_module_conv_dim"),
                        fft_centered=self.fft_centered,
                        fft_normalization=self.fft_normalization,
                        spatial_dims=self.spatial_dims,
                        coil_dim=self.coil_dim - 1,
                        dimensionality=cfg_dict.get("reconstruction_module_dimensionality"),
                        coil_combination_method=self.coil_combination_method,
                    )
                )

            # Keep estimation through the cascades if keep_prediction is True or re-estimate it if False.
            self.reconstruction_module_keep_prediction = cfg_dict.get("reconstruction_module_keep_prediction")

            # initialize weights if not using pretrained cirim
            if not cfg_dict.get("pretrained", False):
                std_init_range = 1 / self.reconstruction_module_recurrent_filters[0] ** 0.5
                self.reconstruction_module.apply(lambda module: rnn_weights_init(module, std_init_range))

            self.dc_weight = torch.nn.Parameter(torch.ones(1))
            self.reconstruction_module_accumulate_predictions = cfg_dict.get(
                "reconstruction_module_accumulate_predictions"
            )

        self.quantitative_maps_scaling_factor = cfg_dict.get("quantitative_maps_scaling_factor")

        quantitative_module_num_cascades = cfg_dict.get("quantitative_module_num_cascades")
        self.quantitative_module = torch.nn.ModuleList(
            [
                qRIMBlock(
                    recurrent_layer=cfg_dict.get("quantitative_module_recurrent_layer"),
                    conv_filters=cfg_dict.get("quantitative_module_conv_filters"),
                    conv_kernels=cfg_dict.get("quantitative_module_conv_kernels"),
                    conv_dilations=cfg_dict.get("quantitative_module_conv_dilations"),
                    conv_bias=cfg_dict.get("quantitative_module_conv_bias"),
                    recurrent_filters=cfg_dict.get("quantitative_module_recurrent_filters"),
                    recurrent_kernels=cfg_dict.get("quantitative_module_recurrent_kernels"),
                    recurrent_dilations=cfg_dict.get("quantitative_module_recurrent_dilations"),
                    recurrent_bias=cfg_dict.get("quantitative_module_recurrent_bias"),
                    depth=cfg_dict.get("quantitative_module_depth"),
                    time_steps=cfg_dict.get("quantitative_module_time_steps"),
                    conv_dim=cfg_dict.get("quantitative_module_conv_dim"),
                    linear_forward_model=SignalForwardModel(
                        sequence=cfg_dict.get("quantitative_module_signal_forward_model_sequence")
                    ),
                    fft_centered=self.fft_centered,
                    fft_normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                    coil_dim=self.coil_dim,
                    coil_combination_method=self.coil_combination_method,
                    dimensionality=quantitative_module_dimensionality,
                )
                for _ in range(quantitative_module_num_cascades)
            ]
        )
        self.quantitative_maps_regularization_factors = cfg_dict.get(
            "quantitative_maps_regularization_factors", [150.0, 150.0, 1000.0, 150.0]
        )

        self.accumulate_predictions = cfg_dict.get("quantitative_module_accumulate_predictions")

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        R2star_map_init: torch.Tensor,
        S0_map_init: torch.Tensor,
        B0_map_init: torch.Tensor,
        phi_map_init: torch.Tensor,
        TEs: List,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        initial_prediction: torch.Tensor,
        anatomy_mask: torch.Tensor,
        sampling_mask: torch.Tensor,
        sigma: float = 1.0,
    ) -> Union[List[List[Tensor]], List[Tensor]]:
        """Forward pass of :class:`qCIRIM`.

        Parameters
        ----------
        R2star_map_init : torch.Tensor
            Initial R2* map of shape [batch_size, n_x, n_y].
        S0_map_init : torch.Tensor
            Initial S0 map of shape [batch_size, n_x, n_y].
        B0_map_init : torch.Tensor
            Initial B0 map of shape [batch_size, n_x, n_y].
        phi_map_init : torch.Tensor
            Initial phase map of shape [batch_size, n_x, n_y].
        TEs : List
            List of echo times.
        y : torch.Tensor
            Subsampled k-space data of shape [batch_size, n_echoes, n_coils, n_x, n_y, 2].
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2].
        initial_prediction : torch.Tensor
            Initial prediction of shape [batch_size, n_x, n_y, 2].
        anatomy_mask : torch.Tensor
            Brain mask of shape [batch_size, 1, n_x, n_y, 1].
        sampling_mask : torch.Tensor
            Sampling mask of shape [batch_size, 1, n_x, n_y, 1].
        sigma : float
            Standard deviation of the noise. Default is ``1.0``.

        Returns
        -------
        List of list of torch.Tensor or torch.Tensor
             If self.accumulate_loss is True, returns a list of all intermediate predictions.
             If False, returns the final estimate.
        """
        if self.use_reconstruction_module:
            cascades_echoes_reconstruction_predictions = []
            sigma = 1.0
            for echo in range(y.shape[1]):
                reconstruction_prediction = y[:, echo, ...].clone()
                initial_reconstruction_prediction_echo = None
                hx = None
                cascades_reconstruction_predictions = []
                for i, cascade in enumerate(self.reconstruction_module):
                    # Forward pass through the cascades
                    reconstruction_prediction, hx = cascade(
                        reconstruction_prediction,
                        y[:, echo, ...],
                        sensitivity_maps,
                        sampling_mask[:, 0, ...],
                        initial_reconstruction_prediction_echo if i == 0 else reconstruction_prediction,
                        hx,
                        sigma,
                        keep_prediction=False if i == 0 else self.reconstruction_module_keep_prediction,
                    )
                    cascades_reconstruction_predictions.append(
                        [torch.view_as_complex(x) for x in reconstruction_prediction]
                    )
                    reconstruction_prediction = reconstruction_prediction[-1]
                cascades_echoes_reconstruction_predictions.append(cascades_reconstruction_predictions)

            # cascades_echoes_reconstruction_predictions is of length n_echoes, len(self.reconstruction_module),
            # self.reconstruction_module_time_steps. We want to concatenate the echoes for each cascade and time step.
            reconstruction_prediction = []
            for cascade in range(len(cascades_echoes_reconstruction_predictions[0])):
                reconstruction_prediction_cascade = []
                for time_step in range(len(cascades_echoes_reconstruction_predictions[0][cascade])):
                    reconstruction_prediction_time_step = []
                    for echo in range(  # pylint: disable=consider-using-enumerate
                        len(cascades_echoes_reconstruction_predictions)
                    ):
                        reconstruction_prediction_time_step.append(
                            cascades_echoes_reconstruction_predictions[echo][cascade][time_step]
                        )
                    reconstruction_prediction_time_step = torch.stack(reconstruction_prediction_time_step, dim=1)
                    if reconstruction_prediction_time_step.shape[-1] != 2:  # type: ignore
                        reconstruction_prediction_time_step = torch.view_as_real(reconstruction_prediction_time_step)
                    reconstruction_prediction_cascade.append(reconstruction_prediction_time_step)
                reconstruction_prediction.append(reconstruction_prediction_cascade)

            final_reconstruction_prediction = reconstruction_prediction[-1][-1]
            if not self.reconstruction_module_accumulate_predictions:
                reconstruction_prediction = final_reconstruction_prediction

            y = fft2(
                coil_combination_method(
                    final_reconstruction_prediction.unsqueeze(self.coil_dim),
                    sensitivity_maps.unsqueeze(self.coil_dim - 1),
                    method=self.coil_combination_method,
                    dim=self.coil_dim - 1,
                ),
                self.fft_centered,
                self.fft_normalization,
                self.spatial_dims,
            )

            R2star_maps_init = []
            S0_maps_init = []
            B0_maps_init = []
            phi_maps_init = []
            for batch_idx in range(final_reconstruction_prediction.shape[0]):
                R2star_map_init, S0_map_init, B0_map_init, phi_map_init = R2star_B0_S0_phi_mapping(
                    final_reconstruction_prediction[batch_idx],
                    TEs,
                    anatomy_mask,
                    scaling_factor=self.quantitative_maps_scaling_factor,
                    fft_centered=self.fft_centered,
                    fft_normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                )
                R2star_maps_init.append(R2star_map_init.squeeze(0))
                S0_maps_init.append(S0_map_init.squeeze(0))
                B0_maps_init.append(B0_map_init.squeeze(0))
                phi_maps_init.append(phi_map_init.squeeze(0))
            R2star_map_init = torch.stack(R2star_maps_init, dim=0).to(y)
            S0_map_init = torch.stack(S0_maps_init, dim=0).to(y)
            B0_map_init = torch.stack(B0_maps_init, dim=0).to(y)
            phi_map_init = torch.stack(phi_maps_init, dim=0).to(y)
        else:
            reconstruction_prediction = initial_prediction.clone()

        R2star_map_init = R2star_map_init / self.quantitative_maps_regularization_factors[0]
        S0_map_init = S0_map_init / self.quantitative_maps_regularization_factors[1]
        B0_map_init = B0_map_init / self.quantitative_maps_regularization_factors[2]
        phi_map_init = phi_map_init / self.quantitative_maps_regularization_factors[3]

        qmaps_prediction = torch.stack([R2star_map_init, S0_map_init, B0_map_init, phi_map_init], dim=1)
        hx = None
        cascades_R2star_maps_prediction = []
        cascades_S0_maps_prediction = []
        cascades_B0_maps_prediction = []
        cascades_phi_maps_prediction = []
        for i, cascade in enumerate(self.quantitative_module):
            # Forward pass through the cascades
            qmaps_prediction, hx = cascade(qmaps_prediction, y, sensitivity_maps, sampling_mask, TEs, hx)
            # Keep the intermediate predictions
            for qmaps_pred in qmaps_prediction:
                cascades_R2star_maps_prediction.append(
                    qmaps_pred[:, 0, ...] * self.quantitative_maps_regularization_factors[0]
                )
                cascades_S0_maps_prediction.append(
                    qmaps_pred[:, 1, ...] * self.quantitative_maps_regularization_factors[1]
                )
                cascades_B0_maps_prediction.append(
                    qmaps_pred[:, 2, ...] * self.quantitative_maps_regularization_factors[2]
                )
                cascades_phi_maps_prediction.append(
                    qmaps_pred[:, 3, ...] * self.quantitative_maps_regularization_factors[3]
                )
            # Keep the final prediction for the next cascade
            qmaps_prediction = qmaps_prediction[-1]

        if not self.accumulate_predictions:
            reconstruction_prediction = (
                reconstruction_prediction[-1][-1] if self.use_reconstruction_module else torch.empty([])
            )
            cascades_R2star_maps_prediction = cascades_R2star_maps_prediction[-1][-1]
            cascades_S0_maps_prediction = cascades_S0_maps_prediction[-1][-1]
            cascades_B0_maps_prediction = cascades_B0_maps_prediction[-1][-1]
            cascades_phi_maps_prediction = cascades_phi_maps_prediction[-1][-1]

        return [
            reconstruction_prediction,
            cascades_R2star_maps_prediction,
            cascades_S0_maps_prediction,
            cascades_B0_maps_prediction,
            cascades_phi_maps_prediction,
        ]

    def process_quantitative_loss(
        self,
        target: torch.Tensor,
        prediction: Union[list, torch.Tensor],
        anatomy_mask: torch.Tensor,
        quantitative_map: str,
        loss_func: torch.nn.Module,
    ) -> torch.Tensor:
        """Processes the quantitative loss.

        Parameters
        ----------
        target : torch.Tensor
            Target data of shape [batch_size, n_x, n_y, 2].
        prediction : Union[list, torch.Tensor]
            Prediction(s) of shape [batch_size, n_x, n_y, 2].
        anatomy_mask : torch.Tensor
            Mask of specified anatomy, e.g. brain. Shape [n_x, n_y].
        quantitative_map : str
            Type of quantitative map to regularize the loss. Must be one of {"R2star", "S0", "B0", "phi"}.
        loss_func : torch.nn.Module
            Loss function. Default is ``torch.nn.L1Loss()``.

        Returns
        -------
        loss: torch.FloatTensor
            If self.accumulate_loss is True, returns an accumulative result of all intermediate losses.
            Otherwise, returns the loss of the last intermediate loss.
        """
        target = torch.abs(self.__abs_output__(target / torch.max(torch.abs(target))))
        anatomy_mask = torch.abs(anatomy_mask).to(target)

        def compute_quantitative_loss(t, p, m):
            p = torch.abs(self.__abs_output__(p / torch.max(torch.abs(p))))

            if "ssim" in str(loss_func).lower():
                return (
                    loss_func(
                        t * m,
                        p * m,
                        data_range=torch.tensor([max(torch.max(t * m).item(), torch.max(p * m).item())])
                        .unsqueeze(dim=0)
                        .to(t),
                    )
                    * self.quantitative_parameters_regularization_factors[quantitative_map]
                )
            return loss_func(t * m, p * m) / self.quantitative_parameters_regularization_factors[quantitative_map]

        if self.accumulate_predictions:
            cascades_loss = []
            for cascade_pred in prediction:
                time_steps_loss = [
                    compute_quantitative_loss(target, time_step_pred, anatomy_mask) for time_step_pred in cascade_pred
                ]
                cascades_loss.append(torch.sum(torch.stack(time_steps_loss, dim=0)) / len(prediction))
            loss = sum(cascades_loss) / len(self.quantitative_module)
        else:
            loss = compute_quantitative_loss(target, prediction, anatomy_mask)
        return loss

    def process_reconstruction_loss(  # noqa: MC0001
        self,
        target: torch.Tensor,
        prediction: Union[list, torch.Tensor],
        sensitivity_maps: torch.Tensor,
        attrs: Dict,
        r: int,
        loss_func: torch.nn.Module,
    ) -> torch.Tensor:
        """Processes the reconstruction loss.

        Parameters
        ----------
        target : torch.Tensor
            Target data of shape [batch_size, n_x, n_y, 2].
        prediction : Union[list, torch.Tensor]
            Prediction(s) of shape [batch_size, n_x, n_y, 2].
        sensitivity_maps : torch.Tensor
            Sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2]. It will be used if self.ssdu is True, to
            expand the target and prediction to multiple coils.
        attrs : Dict
            Attributes of the data with pre normalization values.
        r : int
            The selected acceleration factor.
        loss_func : torch.nn.Module
            Loss function. Must be one of {torch.nn.L1Loss(), torch.nn.MSELoss(),
            atommic.collections.reconstruction.losses.ssim.SSIMLoss()}. Default is ``torch.nn.L1Loss()``.

        Returns
        -------
        loss: torch.FloatTensor
            If self.accumulate_loss is True, returns an accumulative result of all intermediate losses.
            Otherwise, returns the loss of the last intermediate loss.
        """
        if self.unnormalize_loss_inputs:
            for cascade in range(len(prediction)):  # pylint: disable=consider-using-enumerate
                for time_steps in range(len(prediction[cascade])):
                    target, prediction[cascade][time_steps], sensitivity_maps = self.__unnormalize_for_loss_or_log__(
                        target, prediction[cascade][time_steps], sensitivity_maps, attrs, r
                    )

        # If kspace reconstruction loss is used, the target needs to be transformed in k-space.
        if self.kspace_quantitative_loss:
            # If inputs are complex, then they need to be viewed as real.
            if target.shape[-1] != 2 and torch.is_complex(target):
                target = torch.view_as_real(target)

            # Transform to k-space.
            target = fft2(target, self.fft_centered, self.fft_normalization, self.spatial_dims)

            # Ensure loss inputs are both viewed in the same way.
            target = self.__abs_output__(target / torch.max(torch.abs(target)))

            for cascade in range(len(prediction)):  # pylint: disable=consider-using-enumerate
                for time_steps in range(len(prediction[cascade])):
                    if prediction[cascade][time_steps].shape[-1] != 2 and torch.is_complex(
                        prediction[cascade][time_steps]
                    ):
                        prediction[cascade][time_steps] = torch.view_as_real(prediction[cascade][time_steps])
                    prediction[cascade][time_steps] = fft2(
                        prediction[cascade][time_steps],
                        self.fft_centered,
                        self.fft_normalization,
                        self.spatial_dims,
                    )
                    prediction[cascade][time_steps] = self.__abs_output__(
                        prediction[cascade][time_steps] / torch.max(torch.abs(prediction[cascade][time_steps]))
                    )

        elif not self.unnormalize_loss_inputs:
            # Ensure loss inputs are both viewed in the same way.
            target = self.__abs_output__(target / torch.max(torch.abs(target)))
            for cascade in range(len(prediction)):  # pylint: disable=consider-using-enumerate
                for time_steps in range(len(prediction[cascade])):
                    prediction[cascade][time_steps] = self.__abs_output__(
                        prediction[cascade][time_steps] / torch.max(torch.abs(prediction[cascade][time_steps]))
                    )

        def compute_reconstruction_loss(t, p):
            p = torch.abs(p / torch.max(torch.abs(p)))
            t = torch.abs(t / torch.max(torch.abs(t)))
            return torch.mean(torch.tensor([loss_func(t[:, echo], p[:, echo]) for echo in range(t.shape[1])]))

        if self.reconstruction_module_accumulate_predictions:
            cascades_loss = []
            for cascade_prediction in prediction:
                cascade_time_steps_loss = [
                    compute_reconstruction_loss(target, cascade_prediction_time_step_prediction).mean()
                    for cascade_prediction_time_step_prediction in cascade_prediction
                ]
                cascade_loss = [
                    x
                    * torch.logspace(-1, 0, steps=self.reconstruction_module_time_steps).to(cascade_time_steps_loss[0])
                    for x in cascade_time_steps_loss
                ]
                cascades_loss.append(sum(sum(cascade_loss) / len(self.reconstruction_module)))
            loss = sum(cascades_loss) / len(prediction)
        else:
            loss = compute_reconstruction_loss(target, prediction)
        return loss



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import Union

import torch
from torch import nn


class ConvRNNStack(nn.Module):
    """A stack of convolutional RNNs."""

    def __init__(self, convs, rnn):
        """Inits :class:`ConvRNNStack`.

        Parameters
        ----------
        convs : list of torch.nn.Module
            List of convolutional layers.
        rnn : torch.nn.Module
            RNN layer.
        """
        super().__init__()
        self.convs = convs
        self.rnn = rnn

    def forward(self, x: torch.Tensor, hidden: torch.Tensor = None) -> torch.Tensor:
        """Forward p pass of :class:`ConvRNNStack`.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape [batch_size, seq_len, input_size].
        hidden : torch.Tensor
            Initial hidden state of shape [num_layers * num_directions, batch_size, hidden_size].

        Returns
        -------
        torch.Tensor
            Output tensor of shape [batch_size, seq_len, hidden_size].
        """
        return self.rnn(self.convs(x), hidden)


class ConvNonlinear(nn.Module):
    """A convolutional layer with nonlinearity."""

    def __init__(
        self,
        input_size: int,
        features: int,
        conv_dim: int,
        kernel_size: int,
        dilation: int,
        bias: bool,
        nonlinear: Union[str, None] = "ReLU",
    ):
        """Inits :class:`ConvNonlinear`.

        Parameters
        ----------
        input_size : int
            Number of input channels.
        features : int
            Number of output channels.
        conv_dim : int
            Number of dimensions of the convolutional layer.
        kernel_size : int
            Size of the convolutional kernel.
        dilation : int
            Dilation of the convolutional kernel.
        bias : bool
            Whether to use bias.
        nonlinear : str, optional
            Nonlinearity of the convolutional layer. Default is ``"ReLU"``.
        """
        super().__init__()

        self.input_size = input_size
        self.features = features
        self.kernel_size = kernel_size
        self.dilation = dilation
        self.bias = bias
        self.conv_dim = conv_dim
        self.conv_class = self.determine_conv_class(conv_dim)

        if nonlinear is not None and nonlinear.upper() == "RELU":
            self.nonlinear = torch.nn.ReLU()
        elif nonlinear is not None and nonlinear.upper() == "LEAKYRELU":
            self.nonlinear = torch.nn.LeakyReLU()
        elif nonlinear is None:
            self.nonlinear = lambda x: x
        else:
            raise ValueError("Please specify a proper nonlinearity")

        self.padding = [
            torch.nn.ReplicationPad1d(torch.div(dilation * (kernel_size - 1), 2, rounding_mode="trunc").item()),
            torch.nn.ReplicationPad2d(torch.div(dilation * (kernel_size - 1), 2, rounding_mode="trunc").item()),
            torch.nn.ReplicationPad3d(torch.div(dilation * (kernel_size - 1), 2, rounding_mode="trunc").item()),
        ][conv_dim - 1]

        self.conv_layer = self.conv_class(
            in_channels=input_size,
            out_channels=features,
            kernel_size=kernel_size,
            padding=0,  # TODO: check if this is correct
            dilation=dilation,
            bias=bias,
        )

        self.reset_parameters()

    def reset_parameters(self):
        """Resets the parameters of the convolutional layer."""
        torch.nn.init.kaiming_normal_(self.conv_layer.weight, nonlinearity="relu")

        if self.conv_layer.bias is not None:
            nn.init.zeros_(self.conv_layer.bias)

    @staticmethod
    def determine_conv_class(n_dim: int) -> nn.Module:
        """Determines the convolutional layer class."""
        if n_dim == 1:
            return nn.Conv1d
        if n_dim == 2:
            return nn.Conv2d
        if n_dim == 3:
            return nn.Conv3d
        raise ValueError(f"Convolution of: {n_dim} dims is not implemented")

    def extra_repr(self):
        """Extra information about the layer."""
        s = "{input_size}, {features}"
        if "bias" in self.__dict__ and self.bias is not True:
            s += ", bias={bias}"
        if "nonlinear" in self.__dict__ and self.nonlinear != "tanh":
            s += ", nonlinearity={nonlinear}"
        return s.format(**self.__dict__)

    def check_forward_input(self, _input: torch.Tensor) -> torch.Tensor:
        """Checks input for correct size and shape."""
        if _input.size(1) != self.input_size:
            raise RuntimeError(f"input has inconsistent input_size: got {_input.size(1)}, expected {self.input_size}")

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`ConvNonlinear`."""
        return self.nonlinear(self.conv_layer(self.padding(_input)))



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from omegaconf import DictConfig

from atommic.collections.segmentation.nn.dynunet_base.dynunet_block import DynUNet
from atommic.collections.segmentation.nn.segmentationnet import BaseSegmentationNet
from atommic.core.classes.common import typecheck

__all__ = ["SegmentationDYNUNet"]


class SegmentationDYNUNet(BaseSegmentationNet):
    """Implementation of a Dynamic UNet (DynUNet), based on [Isensee2018]_.

    References
    ----------
    .. [Isensee2018] Isensee F, Petersen J, Klein A, Zimmerer D, Jaeger PF, Kohl S, Wasserthal J, Koehler G,
        Norajitra T, Wirkert S, Maier-Hein KH. nnu-net: Self-adapting framework for u-net-based medical image
        segmentation. arXiv preprint arXiv:1809.10486. 2018 Sep 27.

    """

    def build_segmentation_module(self, cfg: DictConfig) -> torch.nn.Module:
        """Build the segmentation module.

        Parameters
        ----------
        cfg : DictConfig
            Configuration object specifying the model's hyperparameters.

        Returns
        -------
        torch.nn.Module
            The segmentation module.
        """
        strides = cfg.get("segmentation_module_strides", (1, 1, 1, 1))
        self.deep_supervision = cfg.get("segmentation_module_deep_supervision", False)
        return DynUNet(
            spatial_dims=cfg.get("dimensionality", 2),
            in_channels=self.input_channels,
            out_channels=cfg.get("segmentation_module_output_channels", 2),
            kernel_size=cfg.get("segmentation_module_kernel_size", 3),
            strides=strides,
            upsample_kernel_size=strides[1:],
            filters=cfg.get("segmentation_module_channels", 64),
            dropout=cfg.get("segmentation_module_dropout", 0.0),
            norm_name=cfg.get("segmentation_module_norm", "instance"),
            act_name=cfg.get("segmentation_module_activation", "leakyrelu"),
            deep_supervision=self.deep_supervision,
            deep_supr_num=cfg.get("segmentation_module_deep_supervision_levels", 1),
        )

    @typecheck()
    def forward(self, image: torch.Tensor, **kwargs) -> torch.Tensor:
        """
        Forward pass of the network.

        Parameters
        ----------
        image : torch.Tensor
            Input image. Shape [batch_size, n_x, n_y] or [batch_size, n_x, n_y, 2]
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        torch.Tensor
            Predicted segmentation. Shape [batch_size, n_classes, n_x, n_y]
        """
        if self.consecutive_slices > 1:
            batch, slices = image.shape[:2]
            image = image.reshape(batch * slices, *image.shape[2:])

        if image.shape[-1] == 2:
            if self.input_channels == 1:
                image = torch.view_as_complex(image).unsqueeze(1)
                if self.magnitude_input:
                    image = torch.abs(image)
            elif self.input_channels == 2 and not self.magnitude_input:
                image = image.permute(0, 3, 1, 2)
            else:
                raise ValueError(f"The input channels must be either 1 or 2. Found: {self.input_channels}")
        elif image.dim() == 3:
            image = image.unsqueeze(1)

        mean = 1.0
        std = 1.0
        if self.normalize:
            image, mean, std = self.norm(image)
        image, pad_sizes = self.pad(image)
        segmentation = self.segmentation_module(image)
        segmentation = self.unpad(segmentation, *pad_sizes)
        if self.normalize:
            segmentation = self.unnorm(segmentation, mean, std)

        if self.deep_supervision and segmentation.dim() == 5:
            # TODO: check if this is correct. They do unbind, but they don't show how they handle the tuples.
            segmentation = torch.sum(segmentation, dim=1)

        if self.normalize_segmentation_output:
            segmentation = (segmentation - segmentation.min()) / (segmentation.max() - segmentation.min())

        if self.consecutive_slices > 1:
            segmentation = segmentation.reshape(batch, slices, *segmentation.shape[1:])

        return torch.abs(segmentation)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.common.data.mri_loader import MRIDataset  # noqa: F401
from atommic.collections.common.data.subsample import (  # noqa: F401
    Equispaced1DMaskFunc,
    Equispaced2DMaskFunc,
    Gaussian1DMaskFunc,
    Gaussian2DMaskFunc,
    MaskFunc,
    Poisson2DMaskFunc,
    Random1DMaskFunc,
    create_masker,
)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/optim/novograd.py

import torch
from torch.optim.optimizer import Optimizer

__all__ = ["Novograd"]


def _check_valid_opt_params(lr, eps, betas):
    """Check if the given learning rate and epsilon are valid."""
    if lr < 0:
        raise ValueError(f"Invalid learning rate: {lr}")
    if eps < 0:
        raise ValueError(f"Invalid epsilon value: {eps}")
    if not (0.0 <= betas[0] < 1.0 and 0.0 <= betas[1] < 1.0):
        raise ValueError(f"Betas have to be between 0 and 1: {betas}")


class Novograd(Optimizer):
    """Implements Novograd algorithm.

    It has been proposed  in "Stochastic Gradient Methods with Layer-wise Adaptive Moments for Training of Deep
    Networks" (https://arxiv.org/abs/1905.11286).
    """

    def __init__(
        self,
        params,
        lr=1e-3,
        betas=(0.95, 0.98),
        eps=1e-8,
        weight_decay=0,
        grad_averaging=False,
        amsgrad=False,
        luc=False,
        luc_trust=1e-3,
        luc_eps=1e-8,
    ):
        """Inits :class:`Novograd`.

        Parameters
        ----------
        params : iterable
            Iterable of parameters to optimize or dicts defining parameter groups.
        lr : float
            Learning rate. Default is ``1e-3``.
        betas : Tuple[float, float], optional
            Coefficients used for computing running averages of gradient and its square. Default is ``(0.9, 0.999)``.
        eps : float, optional
            Term added to the denominator to improve numerical stability. Default is ``1e-8``.
        weight_decay : float, optional
            Weight decay (L2 penalty). Default is ``0``.
        amsgrad : bool, optional
            Whether to use the AMSGrad variant of this algorithm from the paper "On the Convergence of Adam and
            Beyond". Default is ``False``.
        """
        _check_valid_opt_params(lr, eps, betas)
        defaults = {
            "lr": lr,
            "betas": betas,
            "eps": eps,
            "weight_decay": weight_decay,
            "grad_averaging": grad_averaging,
            "amsgrad": amsgrad,
        }
        self.luc = luc
        self.luc_trust = luc_trust
        self.luc_eps = luc_eps
        super().__init__(params, defaults)

    def __setstate__(self, state):
        """Set state for optimizer."""
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault("amsgrad", False)

    def step(self, closure=None):
        """Performs a single optimization step.

        Parameters
        ----------
        closure : str
            A closure that reevaluates the model and returns the loss.

        Returns
        -------
        loss: Loss (if provided)
        """
        loss = closure() if closure is not None else None
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError("Sparse gradients are not supported.")
                amsgrad = group["amsgrad"]
                state = self.state[p]

                # State initialization
                if not state:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros([]).to(state["exp_avg"].device)
                    if amsgrad:
                        # Maintains max of all exp moving avg of squared grad
                        state["max_exp_avg_sq"] = torch.zeros([]).to(state["exp_avg"].device)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                if amsgrad:
                    max_exp_avg_sq = state["max_exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                norm = grad.norm().pow(2)

                if exp_avg_sq == 0:
                    exp_avg_sq.copy_(norm)
                else:
                    exp_avg_sq.mul_(beta2).add_(norm, alpha=1.0 - beta2)

                if amsgrad:
                    # Maintains max of all 2nd moment running avg till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max for normalizing running avg. of gradient
                    denom = max_exp_avg_sq.sqrt().add_(group["eps"])
                else:
                    denom = exp_avg_sq.sqrt().add_(group["eps"])

                grad.div_(denom)
                if group["weight_decay"] != 0:
                    grad.add_(p.data, alpha=group["weight_decay"])
                if group["grad_averaging"]:
                    grad.mul_(1 - beta1)
                exp_avg.mul_(beta1).add_(grad)

                if self.luc:
                    # Clip update so that updates are less than eta*weights
                    data_norm = torch.norm(p.data)
                    grad_norm = torch.norm(exp_avg.data)
                    luc_factor = self.luc_trust * data_norm / (grad_norm + self.luc_eps)
                    luc_factor = min(luc_factor, group["lr"])
                    p.data.add_(exp_avg, alpha=-luc_factor)
                else:
                    p.data.add_(exp_avg, alpha=-group["lr"])

        return loss



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from https://github.com/bduffy0/motion-correction/blob/master/layer/motion_sim.py

import math
import random
from typing import Any, Dict, Optional, Sequence, Tuple

import torch

from atommic.collections.common.parts import utils


def get_center_rect(image: torch.tensor, center_percentage: float = 0.02, dim: int = 0) -> torch.tensor:
    """Get a center rectangle of a given dimension.

    Parameters
    ----------
    image : torch.tensor
        The image to get the center rectangle from.
    center_percentage : float
        The percentage of the image to take as the center rectangle.
    dim : int
        The dimension to take the center rectangle from.

    Returns
    -------
    torch.tensor
        The center rectangle.
    """
    shape = (image[0].item(), image[1].item())
    mask = torch.zeros(shape)
    half_pct = center_percentage / 2
    center = [int(x / 2) for x in shape]
    mask = torch.swapaxes(mask, 0, dim)
    mask[:, center[1] - math.ceil(shape[1] * half_pct) : math.ceil(center[1] + shape[1] * half_pct)] = 1
    mask = torch.swapaxes(mask, 0, dim)
    return mask


def segment_array_by_locs(shape: Sequence[int], locations: Sequence[int]) -> torch.tensor:
    """Generate a segmentation mask based on a list of locations.

    Parameters
    ----------
    shape : Sequence[int]
        The shape of the array to segment.
    locations : Sequence[int]
        The locations to segment the array into.

    Returns
    -------
    torch.tensor
        The segmentation mask.
    """
    mask_out = torch.zeros(torch.prod(shape), dtype=int)
    for i in range(len(locations) - 1):
        loc = [locations[i], locations[i + 1]]
        mask_out[loc[0] : loc[1]] = i + 1
    return mask_out.reshape(shape)


def segments_to_random_indices(shape: Sequence[int], seg_lengths: Sequence[int]) -> torch.tensor:
    """Generate a segmentation mask based on a list of locations.

    Parameters
    ----------
    shape : Sequence[int]
        The shape of the array to segment.
    seg_lengths : Sequence[int]
        The lengths of the segments to generate.

    Returns
    -------
    torch.tensor
        The segmentation mask.
    """
    random_indices = torch.randint(low=0, high=shape, size=(sum(seg_lengths),)).sort()[0]
    seg_mask = torch.zeros(shape).type(torch.int)
    seg_new_indices = torch.cumsum(torch.tensor(seg_lengths), 0).tolist()
    seg_new_indices = [0] + seg_new_indices
    for i in range(len(seg_new_indices) - 1):
        seg_mask[random_indices[seg_new_indices[i] : seg_new_indices[i + 1]]] = i + 1
    return seg_mask


def segments_to_random_blocks(shape: Sequence[int], seg_lengths: Sequence[int]) -> torch.tensor:
    """Generate a segmentation mask based on a list of locations.

    Parameters
    ----------
    shape : Sequence[int]
        The shape of the array to segment.
    seg_lengths : Sequence[int]
        The lengths of the segments to generate.

    Returns
    -------
    torch.tensor
        The segmentation mask.
    """
    seg_mask = torch.zeros(shape).type(torch.int)
    seg_lengths_sorted = sorted(seg_lengths, reverse=True)
    for i, seg_len in enumerate(seg_lengths_sorted):
        loc = torch.randint(low=0, high=seg_mask.size()[0], size=(1,))
        while (sum(seg_mask[loc : loc + seg_len]) != 0) or (loc + seg_len > seg_mask.size()[0]):
            loc = torch.randint(low=0, high=seg_mask.size()[0], size=(1,))
        seg_mask[loc : loc + seg_len] = i + 1
    return seg_mask


def create_rand_partition(im_length: int, num_segments: int):
    """Create a random partition of an array.

    Parameters
    ----------
    im_length : int
        The length of the array to partition.
    num_segments : int
        The number of segments to partition the array into.

    Returns
    -------
    list
        The partition locations.
    """
    rand_segment_locs = sorted(list(torch.randint(im_length, size=(num_segments,))))
    rand_segment_locs[0] = 0
    rand_segment_locs[-1] = None
    return rand_segment_locs


def create_rotation_matrix_3d(angles: Sequence[float]) -> torch.tensor:
    """Create a 3D rotation matrix.

    Parameters
    ----------
    angles : Sequence[float]
        The angles to rotate the matrix by.

    Returns
    -------
    torch.tensor
        The rotation matrix.
    """
    mat1 = torch.FloatTensor(
        [
            [1.0, 0.0, 0.0],
            [0.0, math.cos(angles[0]), math.sin(angles[0])],
            [0.0, -math.sin(angles[0]), math.cos(angles[0])],
        ]
    )
    mat2 = torch.FloatTensor(
        [
            [math.cos(angles[1]), 0.0, -math.sin(angles[1])],
            [0.0, 1.0, 0.0],
            [math.sin(angles[1]), 0.0, math.cos(angles[1])],
        ]
    )
    mat3 = torch.FloatTensor(
        [
            [math.cos(angles[2]), math.sin(angles[2]), 0.0],
            [-math.sin(angles[2]), math.cos(angles[2]), 0.0],
            [0.0, 0.0, 1.0],
        ]
    )
    return (mat1 @ mat2) @ mat3


def translate_kspace(freq_domain: torch.tensor, translations: torch.tensor) -> torch.tensor:
    """Translate a k-space array.

    Parameters
    ----------
    freq_domain : torch.tensor
        The k-space array to translate.
    translations : torch.tensor
        The translations to apply to the k-space array.

    Returns
    -------
    torch.tensor
        The translated k-space array.
    """
    lin_spaces = [torch.linspace(-0.5, 0.5, x) for x in freq_domain.shape[:-1]]
    meshgrids = torch.meshgrid(*lin_spaces, indexing="ij")
    grid_coords = torch.stack([mg.flatten() for mg in meshgrids], 0)
    phase_shift = torch.multiply(grid_coords, translations).sum(axis=0)  # phase shift is added
    exp_phase_shift = torch.exp(-2j * math.pi * phase_shift).to(freq_domain.device)
    motion_kspace = torch.view_as_real(
        torch.multiply(exp_phase_shift, torch.view_as_complex(freq_domain).flatten()).reshape(freq_domain.shape[:-1])
    )

    return motion_kspace


class MotionSimulation:
    """Simulates random translations and rotations in the frequency domain.

    Examples
    --------
    >>> from atommic.collections.motioncorrection.parts import MotionSimulation
    >>> import torch
    >>> motion_simulation = MotionSimulation()
    >>> kspace = torch.randn(1, 1, 256, 256, 2)
    >>> motion_kspace = motion_simulation(kspace)
    >>> motion_kspace.shape
    torch.Size([1, 1, 256, 256, 2])
    """

    def __init__(
        self,
        motion_type: str = "piecewise_transient",
        angle: float = 0,
        translation: float = 10,
        center_percentage: float = 0.02,
        motion_percentage: Sequence[float] = (15, 20),
        num_segments: int = 8,
        random_num_segments: bool = False,
        non_uniform: bool = False,
        spatial_dims: Sequence[int] = (-2, -1),
    ):
        """Inits :class:`MotionSimulation`.

        Parameters
        ----------
        motion_type : str
            The motion_type of motion to simulate.
        angle : float
            The angle to rotate the k-space array by.
        translation : float
            The translation to apply to the k-space array.
        center_percentage : float
            The percentage of the k-space array to center the motion.
        motion_percentage : Sequence[float]
            The percentage of the k-space array to apply the motion.
        num_segments : int
            The number of segments to partition the k-space array into.
        random_num_segments : bool
            Whether to randomly generate the number of segments.
        non_uniform : bool
            Whether to use non-uniform sampling.
        spatial_dims : Sequence[int]
            The spatial dimensions to apply the motion to.
        """
        self.motion_type = motion_type
        self.angle, self.translation = angle, translation
        self.center_percentage = center_percentage

        if motion_percentage[1] == motion_percentage[0]:
            motion_percentage[1] += 1  # type: ignore
        elif motion_percentage[1] < motion_percentage[0]:
            raise ValueError("Uniform is not defined when low>= high.")

        self.motion_percentage = motion_percentage

        self.spatial_dims = spatial_dims
        self._spatial_dims = random.choice(spatial_dims)

        self.num_segments = num_segments
        self.random_num_segments = random_num_segments

        if non_uniform:
            raise NotImplementedError("NUFFT is not implemented. This is a feature to be added in the future.")

        self.trajectory = None
        self.params: Dict[Any, Any] = {}

    def _calc_dimensions(self, shape):
        """Calculate the dimensions to apply the motion to.

        Parameters
        ----------
        shape : Sequence[int]
            The shape of the image.

        Returns
        -------
        Sequence[int]
            The dimensions to apply the motion to.
        """
        pe_dims = [0, 1, 2]
        pe_dims.pop(self._spatial_dims)
        self.phase_encoding_dims = pe_dims
        shape = list(shape)
        if shape[-1] == 2:
            shape = shape[:-1]
        self.shape = shape.copy()
        shape.pop(self._spatial_dims)
        self.phase_encoding_shape = torch.tensor(shape)
        self.num_phase_encoding_steps = self.phase_encoding_shape[0] * self.phase_encoding_shape[1]
        self._spatial_dims = len(self.shape) - 1 if self._spatial_dims == -1 else self._spatial_dims

    def _generate_random_segments(self):
        """Generate random segments."""
        pct_corrupt = torch.distributions.Uniform(*[x / 100 for x in self.motion_percentage]).sample((1, 1))

        corrupt_matrix_shape = torch.tensor([int(x * math.sqrt(pct_corrupt)) for x in self.phase_encoding_shape])

        if torch.prod(corrupt_matrix_shape) == 0:
            corrupt_matrix_shape = [1, 1]

        if self.motion_type in {"gaussian"}:
            num_segments = torch.prod(corrupt_matrix_shape)
        else:
            if not self.random_num_segments:
                num_segments = self.num_segments
            else:
                num_segments = random.randint(1, self.num_segments)

        # segment a smaller vector occupying pct_corrupt percent of the space
        if self.motion_type in {"piecewise_transient", "piecewise_constant"}:
            seg_locs = create_rand_partition(torch.prod(corrupt_matrix_shape), num_segments=num_segments)
        else:
            seg_locs = list(range(num_segments))

        rand_segmentation = segment_array_by_locs(shape=torch.prod(corrupt_matrix_shape), locations=seg_locs)

        seg_lengths = [(rand_segmentation == seg_num).sum() for seg_num in torch.unique(rand_segmentation)]

        # assign segments to a vector with same number of elements as pe-steps
        if self.motion_type in {"piecewise_transient", "gaussian"}:
            seg_vector = segments_to_random_indices(torch.prod(self.phase_encoding_shape), seg_lengths)
        else:
            seg_vector = segments_to_random_blocks(torch.prod(self.phase_encoding_shape), seg_lengths)

        # reshape to phase encoding shape with a random order
        reshape_order = random.choice(["F", "C"])

        if reshape_order == "F":
            seg_array = utils.reshape_fortran(
                seg_vector, (self.phase_encoding_shape[0].item(), self.phase_encoding_shape[1].item())
            )
        else:
            seg_array = seg_vector.reshape((self.phase_encoding_shape[0].item(), self.phase_encoding_shape[1].item()))

        self.order = reshape_order

        # mask center k-space
        mask_not_including_center = (
            get_center_rect(
                self.phase_encoding_shape,
                center_percentage=self.center_percentage,
                dim=1 if reshape_order == "C" else 0,
            )
            == 0
        )

        self.seg_array = seg_array * mask_not_including_center
        self.num_segments = num_segments

    def _get_motion_trajectory(self, translation_rotation=None, random_segments=True):
        """Obtain a motion trajectory.

        Returns
        -------
        torch.tensor
            The random trajectory.
        """

        if random_segments:
            self._generate_random_segments()
        else:
            raise NotImplementedError("Custom segments (masks) not supported")

        if not translation_rotation:
            translations, rotations = self._simulate_random_trajectory()
        else:
            (translations, rotations) = translation_rotation
            if translations.shape[0] != self.num_segments:
                translations = torch.cat((torch.tensor([[0, 0, 0]]), translations), dim=0)
            if rotations.shape[0] != self.num_segments:
                rotations = torch.cat((torch.tensor([[0, 0, 0]]), rotations), dim=0)

        # if segment==0, then no motion
        translations[0, :] = 0
        rotations[0, :] = 0

        # lookup values for each segment
        translations_pe = [translations[:, i][self.seg_array.long()] for i in range(3)]
        rotations_pe = [rotations[:, i][self.seg_array.long()] for i in range(3)]

        # reshape and convert to radians
        translations = torch.stack(
            [torch.broadcast_to(x.unsqueeze(self._spatial_dims), self.shape) for x in translations_pe], 0
        )
        rotations = torch.stack(
            [torch.broadcast_to(x.unsqueeze(self._spatial_dims), self.shape) for x in rotations_pe], 0
        )

        rotations = rotations * (math.pi / 180.0)  # convert to radians

        self.translations = translations.reshape(3, -1)
        self.rotations = rotations.reshape(3, -1).reshape(3, -1)

    def _simulate_random_trajectory(self):
        """Simulate a random trajectory."""
        # generate random translations and rotations
        rand_translations = torch.distributions.normal.Normal(loc=0, scale=self.translation).sample(
            (self.num_segments, 3)
        )
        rand_rotations = torch.distributions.normal.Normal(loc=0, scale=self.angle).sample((self.num_segments, 3))
        return rand_translations, rand_rotations

    def forward(
        self,
        kspace,
        translations_rotations: Optional[Tuple[torch.Tensor, torch.Tensor]] = None,
        apply_backward_transform: bool = False,  # pylint: disable=unused-argument
        apply_forward_transform: bool = False,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`MotionSimulation`.

        Parameters
        ----------
        kspace : torch.Tensor
            The kspace to apply the motion to.
        translations_rotations : Optional[Tuple[torch.Tensor, torch.Tensor]]
            The translations and rotations to apply to the kspace. If None, a random trajectory is generated.
        apply_backward_transform : bool
            Placeholder for the backward transform. Generalizes the Composer, but not used.
        apply_forward_transform : bool
            Placeholder for the forward transform. Generalizes the Composer, but not used.

        Returns
        -------
        torch.Tensor
            The kspace with the motion applied.
        """
        self._calc_dimensions(kspace.shape)
        self._get_motion_trajectory(translations_rotations)

        motion_kspace = translate_kspace(freq_domain=kspace, translations=self.translations)

        return motion_kspace

    def __call__(self, *args, **kwargs):
        """Call :class:`MotionSimulation`."""
        return self.forward(*args, **kwargs)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from einops import rearrange
from torch import einsum, nn

from atommic.collections.reconstruction.nn.unet_base.unet_block import TransposeConvBlock, Unet


class LambdaLayer(nn.Module):
    """Implementation of a Lambda Layer of the Lambda UNet for MRI segmentation, as presented in [Yanglan2021]_.

    References
    ----------
    .. [Yanglan2021] Yanglan Ou, Ye Yuan, Xiaolei Huang, Kelvin Wong, John Volpi, James Z. Wang, Stephen T.C. Wong.
        LambdaUNet: 2.5D Stroke Lesion Segmentation of Diffusion-weighted MR Images. 2021.
        https://arxiv.org/abs/2104.13917
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        query_depth: int = 16,
        intra_depth: int = 1,
        receptive_kernel: int = 3,
        temporal_kernel: int = 1,
        heads: int = 4,
        num_slices: int = 1,
    ):
        """Inits :class:`LambdaLayer`.

        Parameters
        ----------
        in_channels : int
            Number of _input channels.
        out_channels : int
            Number of output channels.
        query_depth : int, optional
            Number of channels for the keys. Default is ``16``.
        intra_depth : int, optional
            Number of neighboring slices. Default is ``1``.
        receptive_kernel : int, optional
            Local context kernel size. Default is ``3``.
        temporal_kernel : int, optional
            Temporal kernel. Default is ``1``.
        heads : int, optional
            Number of query heads. Default is ``4``.
        num_slices : int, optional
            Number of slices. Default is ``1``.
        """
        super().__init__()
        self.dim_in = in_channels
        self.dim_out = out_channels

        self.q_depth = query_depth
        self.intra_depth = intra_depth

        if (out_channels % heads) != 0:
            raise AssertionError("out_channels must be divisible by number of heads for multi-head query.")
        self.v_depth = out_channels // heads
        self.heads = heads

        self.num_slices = num_slices

        self.receptive_kernel = receptive_kernel
        self.temporal_kernel = temporal_kernel

        self.to_q = nn.Sequential(nn.Conv2d(in_channels, query_depth * heads, kernel_size=1, bias=False))
        self.to_k = nn.Sequential(nn.Conv2d(in_channels, query_depth * intra_depth, kernel_size=1, bias=False))
        self.to_v = nn.Sequential(nn.Conv2d(in_channels, self.v_depth * intra_depth, kernel_size=1, bias=False))

        if (receptive_kernel % 2) != 1:
            raise AssertionError("Receptive kernel size should be odd.")
        self.pos_conv = nn.Conv3d(
            intra_depth,
            query_depth,
            (1, receptive_kernel, receptive_kernel),
            padding=(0, receptive_kernel // 2, receptive_kernel // 2),
        )

        if temporal_kernel >= 3:
            if temporal_kernel > num_slices:
                raise AssertionError
            if (temporal_kernel % 2) != 1:
                raise AssertionError("Temporal kernel size should be odd.")
            self.temp_conv = nn.Conv2d(
                intra_depth,
                query_depth,
                (1, temporal_kernel),
                padding=(0, temporal_kernel // 2),
            )

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`LambdaLayer`."""
        _, _, height, width = (*_input.shape,)  # type: ignore

        q = self.to_q(_input)
        k = self.to_k(_input)
        v = self.to_v(_input)

        q = rearrange(q, "b (h k) hh ww -> b h k (hh ww)", h=self.heads)
        k = rearrange(k, "b (u k) hh ww -> b u k (hh ww)", u=self.intra_depth)
        v = rearrange(v, "b (u v) hh ww -> b u v (hh ww)", u=self.intra_depth)

        k = k.softmax(dim=-1)

        lc = einsum("b u k m, b u v m -> b k v", k, v)
        Yc = einsum("b h k n, b k v -> b h v n", q, lc)

        v_p = rearrange(v, "b u v (hh ww) -> b u v hh ww", hh=height, ww=width)  # type: ignore
        lp = self.pos_conv(v_p)
        Yp = einsum("b h k n, b k v n -> b h v n", q, lp.flatten(3))

        if self.temporal_kernel >= 3:
            v_t = rearrange(v, "(g t) u v p -> (g p) u v t", t=self.num_slices)
            lt = self.temp_conv(v_t)
            lt = rearrange(lt, "(g p) k v t -> (g t) k v p", p=height * width)  # type: ignore
            Yt = einsum("b h k n, b k v n -> b h v n", q, lt)
            Y = Yc + Yp + Yt
        else:
            Y = Yc + Yp

        return rearrange(Y, "b h v (hh ww) -> b (h v) hh ww", hh=height, ww=width)  # type: ignore


class LambdaBlock(nn.Module):
    """Implementation of a Lambda Black of the Lambda UNet for MRI segmentation, as presented in [Yanglan2021]_.

    References
    ----------
    .. [Yanglan2021] Yanglan Ou, Ye Yuan, Xiaolei Huang, Kelvin Wong, John Volpi, James Z. Wang, Stephen T.C. Wong.
        LambdaUNet: 2.5D Stroke Lesion Segmentation of Diffusion-weighted MR Images. 2021.
        https://arxiv.org/abs/2104.13917
    """

    def __init__(
        self,
        in_chans: int,
        out_chans: int,
        drop_prob: float,
        query_depth: int = 16,
        intra_depth: int = 4,
        receptive_kernel: int = 3,
        temporal_kernel: int = 1,
        num_slices: int = 1,
    ):
        """Inits :class:`LambdaBlock`.

        Parameters
        ----------
        in_chans : int
            Number of _input channels.
        out_chans : int
            Number of output channels.
        drop_prob : float
            Dropout probability.
        query_depth : int, optional
            Number of channels for the keys. Default is ``16``.
        intra_depth : int, optional
            Number of neighboring slices. Default is ``4``.
        receptive_kernel : int, optional
            Local context kernel size. Default is ``3``.
        temporal_kernel : int, optional
            Temporal kernel. Default is ``1``.
        num_slices : int, optional
            Number of slices. Default is ``1``.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans
        self.drop_prob = drop_prob

        self.layers = nn.Sequential(
            LambdaLayer(
                in_chans,
                out_chans,
                query_depth=query_depth,
                intra_depth=intra_depth,
                receptive_kernel=receptive_kernel,
                temporal_kernel=temporal_kernel,
                heads=max(1, out_chans // 32),
                num_slices=num_slices,
            ),
            nn.InstanceNorm2d(out_chans),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
            nn.Dropout2d(drop_prob),
            LambdaLayer(
                out_chans,
                out_chans,
                query_depth=query_depth,
                intra_depth=intra_depth,
                receptive_kernel=receptive_kernel,
                temporal_kernel=temporal_kernel,
                heads=max(1, out_chans // 32),
                num_slices=num_slices,
            ),
            nn.InstanceNorm2d(out_chans),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
            nn.Dropout2d(drop_prob),
        )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`LambdaBlock`."""
        return self.layers(image)


class LambdaUNet(Unet):
    """Implementation of an extended UNet with Lambda blocks, as presented in [Yanglan2021]_.

    References
    ----------
    .. [Yanglan2021] Yanglan Ou, Ye Yuan, Xiaolei Huang, Kelvin Wong, John Volpi, James Z. Wang, Stephen T.C. Wong.
        LambdaUNet: 2.5D Stroke Lesion Segmentation of Diffusion-weighted MR Images. 2021.
        https://arxiv.org/abs/2104.13917
    """

    def __init__(
        self,
        in_chans: int,
        out_chans: int,
        chans: int = 32,
        num_pool_layers: int = 4,
        drop_prob: float = 0.0,
        query_depth: int = 16,
        intra_depth: int = 4,
        receptive_kernel: int = 3,
        temporal_kernel: int = 1,
        num_slices: int = 1,
    ):
        """Inits :class:`LambdaUNet`.

        Parameters
        ----------
        in_chans : int
            Number of _input channels.
        out_chans : int
            Number of output channels.
        chans : int, optional
            Number of channels. Default is ``32``.
        num_pool_layers : int, optional
            Number of pooling layers. Default is ``4``.
        drop_prob : float
            Dropout probability. Default is ``0.0``.
        query_depth : int, optional
            Number of channels for the keys. Default is ``16``.
        intra_depth : int, optional
            Number of neighboring slices. Default is ``4``.
        receptive_kernel : int, optional
            Local context kernel size. Default is ``3``.
        temporal_kernel : int, optional
            Temporal kernel. Default is ``1``.
        num_slices : int, optional
            Number of slices. Default is ``1``.
        """
        super().__init__(
            in_chans=in_chans, out_chans=out_chans, chans=chans, num_pool_layers=num_pool_layers, drop_prob=drop_prob
        )

        self.in_chans = in_chans
        self.out_chans = out_chans
        self.chans = chans
        self.num_pool_layers = num_pool_layers
        self.drop_prob = drop_prob

        self.down_sample_layers = nn.ModuleList(
            [
                LambdaBlock(
                    in_chans,
                    chans,
                    drop_prob=drop_prob,
                    query_depth=query_depth,
                    intra_depth=intra_depth,
                    receptive_kernel=receptive_kernel,
                    temporal_kernel=temporal_kernel,
                    num_slices=num_slices,
                )
            ]
        )
        ch = chans
        for _ in range(num_pool_layers - 1):
            self.down_sample_layers.append(
                LambdaBlock(
                    ch,
                    ch * 2,
                    drop_prob=drop_prob,
                    query_depth=query_depth,
                    intra_depth=intra_depth,
                    receptive_kernel=receptive_kernel,
                    temporal_kernel=temporal_kernel,
                    num_slices=num_slices,
                )
            )
            ch = ch * 2
        self.conv = LambdaBlock(
            ch,
            ch * 2,
            drop_prob=drop_prob,
            query_depth=query_depth,
            intra_depth=intra_depth,
            receptive_kernel=receptive_kernel,
            temporal_kernel=temporal_kernel,
            num_slices=num_slices,
        )

        self.up_conv = nn.ModuleList()
        self.up_transpose_conv = nn.ModuleList()
        for _ in range(num_pool_layers - 1):
            self.up_transpose_conv.append(TransposeConvBlock(ch * 2, ch))
            self.up_conv.append(
                LambdaBlock(
                    ch * 2,
                    ch,
                    drop_prob=drop_prob,
                    query_depth=query_depth,
                    intra_depth=intra_depth,
                    receptive_kernel=receptive_kernel,
                    temporal_kernel=temporal_kernel,
                    num_slices=num_slices,
                )
            )
            ch = ch // 2

        self.up_transpose_conv.append(TransposeConvBlock(ch * 2, ch))
        self.up_conv.append(
            nn.Sequential(
                LambdaBlock(
                    ch * 2,
                    ch,
                    drop_prob=drop_prob,
                    query_depth=query_depth,
                    intra_depth=intra_depth,
                    receptive_kernel=receptive_kernel,
                    temporal_kernel=temporal_kernel,
                    num_slices=num_slices,
                ),
                nn.Conv2d(ch, self.out_chans, kernel_size=1, stride=1),
            )
        )



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from:
# https://github.com/Project-MONAI/MONAI/blob/c38d503a587f1779914bd071a1b2d66a6d9080c2/monai/networks/layers/weight_init.py#L45

import math
from typing import Optional, Sequence, Tuple, Union

import numpy as np
import torch
from torch import nn


class Convolution(nn.Sequential):
    """Constructs a convolution with normalization, optional dropout, and optional activation layers::
        -- (Conv|ConvTrans) -- (Norm -- Dropout -- Acti) --
    if ``conv_only`` set to ``True``::
        -- (Conv|ConvTrans) --

    .. note::
        This is a wrapper for monai implementation. See:
        https://github.com/Project-MONAI/MONAI/blob/c38d503a587f1779914bd071a1b2d66a6d9080c2/monai/networks/layers/\
        weight_init.py
    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        strides: Union[Sequence[int], int] = 1,
        kernel_size: Union[Sequence[int], int] = 3,
        adn_ordering: str = "NDA",  # pylint: disable=unused-argument
        act: Optional[Union[Tuple, str]] = "PRELU",  # pylint: disable=unused-argument
        norm: Optional[Union[Tuple, str]] = "INSTANCE",  # pylint: disable=unused-argument
        dropout: Optional[Union[Tuple, str, float]] = None,  # pylint: disable=unused-argument
        dropout_dim: Optional[int] = 1,  # pylint: disable=unused-argument
        dilation: Union[Sequence[int], int] = 1,
        groups: int = 1,
        bias: bool = True,
        conv_only: bool = False,  # pylint: disable=unused-argument
        is_transposed: bool = False,
        padding: Optional[Union[Sequence[int], int]] = None,
        output_padding: Optional[Union[Sequence[int], int]] = None,
    ):
        """Inits :class:`Convolution`.

        Parameters
        ----------
        spatial_dims : int
            Number of spatial dimensions.
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        kernel_size : Union[Sequence[int], int]
            Size of the convolving kernel.
        stride : Union[Sequence[int], int], optional
            Stride of the convolution. Default is ``1``.
        kernel_size : Union[Sequence[int], int]
            Size of the convolving kernel. Default is ``3``.
        adn_ordering : str, optional
            A string representing the ordering of activation, normalization, and dropout. Default is ``"NDA"``.
        act : Union[Type[nn.Module], Tuple[Type[nn.Module], dict]], optional
            Activation type and arguments. Default is ``PReLU``.
        norm : Union[Type[nn.Module], Tuple[Type[nn.Module], dict]], optional
            Feature normalization type and arguments. Default is ``instance norm``.
        dropout : float, optional
            Dropout ratio. Default is ``no dropout``.
        dropout_dim : int, optional
            Determine the spatial dimensions of dropout. Default is ``1``. \
            - When dropout_dim = 1, randomly zeroes some of the elements for each channel.
            - When dropout_dim = 2, Randomly zeroes out entire channels (a channel is a 2D feature map).
            - When dropout_dim = 3, Randomly zeroes out entire channels (a channel is a 3D feature map).
            The value of dropout_dim should be no larger than the value of `spatial_dims`.
        dilation : int, optional
            Dilation rate. Default is ``1``.
        groups : int, optional
            Controls the connections between inputs and outputs. Default is ``1``.
        bias : bool, optional
            Whether to have a bias term. Default is ``True``.
        conv_only : bool, optional
            Whether to use the convolutional layer only. Default is ``False``.
        is_transposed : bool, optional
            If ``True`` uses ConvTrans instead of Conv. Default is ``False``.
        padding : Union[Sequence[int], int], optional
            Controls the amount of implicit zero-paddings on both sides for padding number of points for each
            dimension. Default is ``None``.
        output_padding : Union[Sequence[int], int], optional
            Controls the additional size added to one side of the output shape. Default is ``None``.
        """
        super().__init__()
        self.spatial_dims = spatial_dims
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.is_transposed = is_transposed
        if padding is None:
            padding = same_padding(kernel_size, dilation)

        if self.spatial_dims == 1:
            if is_transposed:
                conv_type = nn.ConvTranspose1d
            else:
                conv_type = nn.Conv1d
        elif self.spatial_dims == 2:
            if is_transposed:
                conv_type = nn.ConvTranspose2d
            else:
                conv_type = nn.Conv2d
        elif self.spatial_dims == 3:
            if is_transposed:
                conv_type = nn.ConvTranspose3d
            else:
                conv_type = nn.Conv3d
        else:
            raise ValueError(f"Unsupported spatial_dims: {self.spatial_dims}")

        conv: nn.Module
        if is_transposed:
            if output_padding is None:
                output_padding = stride_minus_kernel_padding(1, strides)
            conv = conv_type(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=strides,
                padding=padding,
                output_padding=output_padding,
                groups=groups,
                bias=bias,
                dilation=dilation,
            )
        else:
            conv = conv_type(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=strides,
                padding=padding,
                dilation=dilation,
                groups=groups,
                bias=bias,
            )

        self.add_module("conv", conv)


def stride_minus_kernel_padding(
    kernel_size: Union[Sequence[int], int], stride: Union[Sequence[int], int]
) -> Union[Tuple[int, ...], int]:
    """Calculate the output padding for the given kernel size and stride.

    Parameters
    ----------
    kernel_size : Union[Sequence[int], int]
        The kernel size.
    stride : Union[Sequence[int], int]
        The stride.

    Returns
    -------
    Union[Tuple[int, ...], int]
        The output padding.
    """
    kernel_size_np = np.atleast_1d(kernel_size)
    stride_np = np.atleast_1d(stride)

    out_padding_np = stride_np - kernel_size_np
    out_padding = tuple(int(p) for p in out_padding_np)

    return out_padding if len(out_padding) > 1 else out_padding[0]


def get_padding(
    kernel_size: Union[Sequence[int], int], stride: Union[Sequence[int], int]
) -> Union[Tuple[int, ...], int]:
    """Calculate the padding for the given kernel size and stride.

    Parameters
    ----------
    kernel_size : Union[Sequence[int], int]
        The kernel size.
    stride : Union[Sequence[int], int]
        The stride.

    Returns
    -------
    Union[Tuple[int, ...], int]
        The padding.
    """
    kernel_size_np = np.atleast_1d(kernel_size)
    stride_np = np.atleast_1d(stride)
    padding_np = (kernel_size_np - stride_np + 1) / 2
    if np.min(padding_np) < 0:
        raise AssertionError("padding value should not be negative, please change the kernel size and/or stride.")
    padding = tuple(int(p) for p in padding_np)

    return padding if len(padding) > 1 else padding[0]


def get_output_padding(
    kernel_size: Union[Sequence[int], int], stride: Union[Sequence[int], int], padding: Union[Sequence[int], int]
) -> Union[Tuple[int, ...], int]:
    """Calculate the output padding for the given kernel size, stride and padding.

    Parameters
    ----------
    kernel_size : Union[Sequence[int], int]
        The kernel size.
    stride : Union[Sequence[int], int]
        The stride.

    Returns
    -------
    Union[Tuple[int, ...], int]
        The output padding.
    """
    kernel_size_np = np.atleast_1d(kernel_size)
    stride_np = np.atleast_1d(stride)
    padding_np = np.atleast_1d(padding)

    out_padding_np = 2 * padding_np + stride_np - kernel_size_np
    if np.min(out_padding_np) < 0:
        raise AssertionError("out_padding value should not be negative, please change the kernel size and/or stride.")
    out_padding = tuple(int(p) for p in out_padding_np)

    return out_padding if len(out_padding) > 1 else out_padding[0]


def same_padding(
    kernel_size: Union[Sequence[int], int], dilation: Union[Sequence[int], int] = 1
) -> Union[Tuple[int, ...], int]:
    """Return the padding value needed to ensure a convolution using the given kernel size produces an output of the
    same shape as the input for a stride of 1, otherwise ensure a shape of the input divided by the stride rounded
    down.

    Raises
    ------
        NotImplementedError: When ``np.any((kernel_size - 1) * dilation % 2 == 1)``.

    Parameters
    ----------
    kernel_size : Union[Sequence[int], int]
        The kernel size.
    dilation : Union[Sequence[int], int]
        The dilation.

    Returns
    -------
    Union[Tuple[int, ...], int]
        The padding.
    """
    kernel_size_np = np.atleast_1d(kernel_size)
    dilation_np = np.atleast_1d(dilation)

    if np.any((kernel_size_np - 1) * dilation % 2 == 1):
        raise NotImplementedError(
            f"Same padding not available for kernel_size={kernel_size_np} and dilation={dilation_np}."
        )

    padding_np = (kernel_size_np - 1) / 2 * dilation_np
    padding = tuple(int(p) for p in padding_np)

    return padding if len(padding) > 1 else padding[0]


def get_conv_layer(
    spatial_dims: int,
    in_channels: int,
    out_channels: int,
    kernel_size: Union[Sequence[int], int] = 3,
    stride: Union[Sequence[int], int] = 1,
    act: Optional[Union[Tuple, str]] = nn.PReLU,
    norm: Optional[Union[Tuple, str]] = nn.InstanceNorm2d,
    dropout: Optional[Union[Tuple, str, float]] = None,
    bias: bool = False,
    conv_only: bool = True,
    is_transposed: bool = False,
) -> Convolution:
    """Get a convolution layer with the given parameters.

    Parameters
    ----------
    spatial_dims : int
        The number of spatial dimensions.
    in_channels : int
        The number of input channels.
    out_channels : int
        The number of output channels.
    kernel_size : Union[Sequence[int], int]
        The kernel size. Default is ``3``.
    stride : Union[Sequence[int], int]
        The stride. Default is ``1``.
    act : Optional[Union[Tuple, str]]
        The activation function. Default is ``nn.PReLU``.
    norm : Optional[Union[Tuple, str]]
        The normalization layer. Default is ``nn.InstanceNorm2d``.
    dropout : Optional[Union[Tuple, str, float]]
        The dropout layer. Default is ``None``.
    bias : bool
        Whether to add a bias. Default is ``False``.
    conv_only : bool
        Whether to only return the convolution layer. Default is ``True``.
    is_transposed : bool
        Whether to use a transposed convolution. Default is ``False``.

    Returns
    -------
    Convolution
        The convolution layer.
    """
    padding = get_padding(kernel_size, stride)
    output_padding = None
    if is_transposed:
        output_padding = get_output_padding(kernel_size, stride, padding)
    return Convolution(
        spatial_dims,
        in_channels,
        out_channels,
        strides=stride,
        kernel_size=kernel_size,
        act=act,
        norm=norm,
        dropout=dropout,
        bias=bias,
        conv_only=conv_only,
        is_transposed=is_transposed,
        padding=padding,
        output_padding=output_padding,
    )


def _no_grad_trunc_normal_(tensor: torch.Tensor, mean: float = 0.0, std: float = 1.0, a: float = -2.0, b: float = 2.0):
    """Tensor initialization with truncated normal distribution.

    Based on:
    https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    https://github.com/rwightman/pytorch-image-models

    Parameters
    ----------
    tensor : torch.Tensor
        The tensor to initialize.
    mean : float
        The mean of the normal distribution. Default is ``0.0``.
    std : float
        The standard deviation of the normal distribution. Default is ``1.0``.
    a : float
        The lower bound of the truncated normal distribution. Default is ``-2.0``.
    b : float
        The upper bound of the truncated normal distribution. Default is ``2.0``.
    """

    def norm_cdf(x: float) -> float:
        """Normal cumulative distribution function."""
        return (1.0 + math.erf(x / math.sqrt(2.0))) / 2.0

    with torch.no_grad():
        tensor.uniform_(2 * norm_cdf((a - mean) / std) - 1, 2 * norm_cdf((b - mean) / std) - 1)
        tensor.erfinv_()
        tensor.mul_(std * math.sqrt(2.0))
        tensor.add_(mean)
        tensor.clamp_(min=a, max=b)
        return tensor


def trunc_normal_(
    tensor: torch.Tensor, mean: float = 0.0, std: float = 1.0, a: float = -2.0, b: float = 2.0
) -> torch.Tensor:
    """Tensor initialization with truncated normal distribution.

    Based on:
    https://github.com/rwightman/pytorch-image-models

    Parameters
    ----------
    tensor : torch.Tensor
        The tensor to initialize.
    mean : float
        The mean of the normal distribution. Default is ``0.0``.
    std : float
        The standard deviation of the normal distribution. Default is ``1.0``.
    a : float
        The lower bound of the truncated normal distribution. Default is ``-2.0``.
    b : float
        The upper bound of the truncated normal distribution. Default is ``2.0``.
    """
    if std <= 0:
        raise ValueError("the standard deviation should be greater than zero.")
    if a >= b:
        raise ValueError("minimum cutoff value (a) should be smaller than maximum cutoff value (b).")
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.multitask.rs.nn.idslr import IDSLR  # noqa: F401
from atommic.collections.multitask.rs.nn.idslr_unet import IDSLRUNet  # noqa: F401
from atommic.collections.multitask.rs.nn.mtlrs import MTLRS  # noqa: F401
from atommic.collections.multitask.rs.nn.mtlrs_base.mtlrs_block import MTLRSBlock  # noqa: F401
from atommic.collections.multitask.rs.nn.recseg_unet import RecSegUNet  # noqa: F401
from atommic.collections.multitask.rs.nn.segnet import SegNet  # noqa: F401
from atommic.collections.multitask.rs.nn.seranet import SERANet  # noqa: F401



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import warnings
from typing import Union

import numpy as np
import torch
from runstats import Statistics
from scipy.ndimage import _ni_support, binary_erosion, distance_transform_edt, generate_binary_structure
from scipy.spatial.distance import directed_hausdorff
from torchmetrics import functional as F

from atommic.collections.segmentation.losses import Dice
from atommic.collections.segmentation.losses.dice import one_hot
from atommic.collections.segmentation.losses.utils import do_metric_reduction


def asd(x, y, voxelspacing=None, connectivity=1):
    """Compute Average Symmetric Surface Distance (ASD) between a binary object and its reference.

    Parameters
    ----------
    x : torch.Tensor
        Ground Truth Tensor.
    y : torch.Tensor
        Prediction Tensor.
    voxelspacing : Voxel Spacing. Defaults to ``None``.
    connectivity : int
        Connectivity. Defaults to ``1``.

    Returns
    -------
    float
        Average Symmetric Surface Distance (ASD) between a binary object and its reference.

    Examples
    --------
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import asd
    >>> import torch
    >>> datax = torch.randint(0, 2, (3, 2, 100, 100))
    >>> datay = torch.randint(0, 2, (3, 2, 100, 100))
    >>> asd(datax, datay)
    0.5010349308997433
    """
    sd1 = np.mean(surface_distances(y, x, voxelspacing, connectivity))  # pylint: disable=arguments-out-of-order
    sd2 = np.mean(surface_distances(x, y, voxelspacing, connectivity))
    return (sd1 + sd2) / 2.0


def binary_cross_entropy_with_logits_metric(x: torch.Tensor, y: torch.Tensor, reduction: str = "mean") -> float:
    """Compute Binary Cross Entropy with Logits.

    Parameters
    ----------
    x : torch.Tensor
        Ground Truth Tensor.
    y : torch.Tensor
        Prediction Tensor.
    reduction : str
        Specifies the reduction to apply to the output: ``none`` | ``mean`` | ``sum``.
        ``none``: no reduction will be applied. ``mean``: the sum of the output will be divided by the number of
        elements. ``sum``: the output will be summed. Default is``mean``.

    Returns
    -------
    float
        Binary Cross Entropy with Logits.

    Examples
    --------
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import \
    binary_cross_entropy_with_logits_metric
    >>> import torch
    >>> datax = torch.randint(0, 2, (3, 2, 100, 100))
    >>> datay = torch.randint(0, 2, (3, 2, 100, 100))
    >>> binary_cross_entropy_with_logits_metric(datax, datay)
    0.7518648505210876

    .. note::
        This function is equivalent to `torch.nn.functional.binary_cross_entropy_with_logits` with `reduction='mean'`.
        Source: https://pytorch.org/docs/stable/generated/torch.nn.functional.binary_cross_entropy_with_logits.html
    """
    if isinstance(x, np.ndarray):
        x = torch.from_numpy(x)
    if isinstance(y, np.ndarray):
        y = torch.from_numpy(y)
    return torch.nn.functional.binary_cross_entropy_with_logits(x.float(), y.float(), reduction=reduction).item()


def dice_metric(
    x: torch.Tensor,
    y: torch.Tensor,
    include_background: bool = True,
    to_onehot_y: bool = False,
    sigmoid: bool = False,
    softmax: bool = False,
    other_act: Union[str, None] = None,
    squared_y: bool = False,
    jaccard: bool = False,
    flatten: bool = False,
    reduction: Union[str, None] = "mean_batch",
    smooth_nr: float = 1e-5,
    smooth_dr: float = 1e-5,
    batch: bool = True,
) -> float:
    """Compute Dice Score.

    Parameters
    ----------
    x : torch.Tensor
        Ground Truth Tensor.
    y : torch.Tensor
        Prediction Tensor.
    include_background : bool
        Whether to skip Dice computation on the first channel of the predicted output. Default is ``True``.
    to_onehot_y : bool
        Whether to convert `y` into the one-hot format. Default is ``False``.
    sigmoid : bool
        Whether to add sigmoid function to the input data. Default is ``True``.
    softmax : bool
        Whether to add softmax function to the input data. Default is ``False``.
    other_act : Union[str, None]
        Use this parameter if you want to apply another type of activation layer. Default is ``None``.
    squared_y : bool
        Whether to square the prediction before calculating Dice. Default is ``False``.
    jaccard : bool
        Whether to compute Jaccard Index as a loss. Default is ``False``.
    flatten : bool
        Whether to flatten input data. Default is ``False``.
    reduction : Union[str, None]
        Specifies the reduction to apply to the output: ``none`` | ``mean`` | ``sum``.
        ``none``: no reduction will be applied. ``mean``: the sum of the output will be divided by the number of
        elements. ``sum``: the output will be summed. Default is ``mean``.
    smooth_nr : float
        A small constant added to the numerator to avoid ``nan`` when all items are 0.
    smooth_dr : float
        A small constant added to the denominator to avoid ``nan`` when all items are 0.
    batch : bool
        If True, compute Dice loss for each batch and return a tensor with shape (batch_size,).
        If False, compute Dice loss for the whole batch and return a tensor with shape (1,).
        Default is ``True``.
    Returns
    -------
    float
        Dice Score.

    Examples
    --------
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import dice_metric
    >>> import torch
    >>> datax = torch.randint(0, 2, (3, 2, 100, 100))
    >>> datay = torch.randint(0, 2, (3, 2, 100, 100))
    >>> dice_metric(datax, datay)
    0.5016108751296997
    """
    custom_dice = Dice(
        include_background=include_background,
        to_onehot_y=to_onehot_y,
        sigmoid=sigmoid,
        softmax=softmax,
        other_act=other_act,  # type: ignore
        squared_pred=squared_y,
        jaccard=jaccard,
        flatten=flatten,
        reduction=reduction,  # type: ignore
        smooth_nr=smooth_nr,
        smooth_dr=smooth_dr,
        batch=batch,
    )
    dice_score, _ = custom_dice(x, y)
    return dice_score.item()


def f1_per_class_metric(
    x: torch.Tensor,
    y: torch.Tensor,
    beta: float = 1e-5,
    average: str = "mean",
    mdmc_average: str = "samplewise",
    threshold: float = 0.0,
) -> float:
    """Compute F1 Score per Class. If the input has only one class, the output will be a list with one element.

    Parameters
    ----------
    x : torch.Tensor
        Ground Truth Tensor.
    y : torch.Tensor
        Prediction Tensor.
    beta : float
        Beta value for F1 score. Default is ``1e-5``.
    average : str
        Defines the averaging performed in the binary case:
        ``micro`` calculates metrics globally,
        ``macro`` calculates metrics for each label, and finds their unweighted mean,
        ``weighted`` calculates metrics for each label, and finds their average, weighted by support
        (the number of true instances for each label),
        ``none`` returns the score for each class.
        Default is ``none``.
    mdmc_average : str
        Defines the averaging performed in the multiclass case:
        ``samplewise`` calculates metrics for each sample, and finds their unweighted mean,
        ``global`` calculates metrics globally, across all samples.
        Default is ``samplewise``.
    threshold : float
        Threshold value for binarization. Default is ``0.0``.

    Returns
    -------
    float
        F1 Score per Class.

    Examples
    --------
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import f1_per_class_metric
    >>> import torch
    >>> datax = torch.randint(0, 2, (3, 2, 100, 100))
    >>> datay = torch.randint(0, 2, (3, 2, 100, 100))
    >>> f1_per_class_metric(datax, datay)
    [0.49855247139930725, 0.49478909373283386]

    .. note::
        This function is a wrapper for `torchmetrics.functional.classification.fbeta_score`.
    """
    if isinstance(x, np.ndarray):
        x = torch.from_numpy(x)
    if isinstance(y, np.ndarray):
        y = torch.from_numpy(y)

    f1_per_class = F.fbeta_score(
        y.to(torch.uint8),
        x.to(torch.uint8),
        task="binary",
        beta=beta,
        average=average,
        multidim_average=mdmc_average,
        num_classes=x.shape[1],
        threshold=threshold,
    )
    if f1_per_class.dim() == 0:
        f1_per_class = torch.stack([f1_per_class] * x.shape[1])
    return f1_per_class.mean()


def hausdorff_distance_metric(x: torch.Tensor, y: torch.Tensor, batched: bool = True, sum_method='max') -> float:
    """Compute Hausdorff Distance.

    The Hausdorff distance is computed as the maximum between x to y and y to x.

    Parameters
    ----------
    x : torch.Tensor
        Ground Truth Tensor.
    y : torch.Tensor
        Prediction Tensor.
    batched : bool
        If True, compute Hausdorff distance for each batch and return a tensor with shape (batch_size,).
        If False, compute Hausdorff distance for the whole batch and return a tensor with shape (1,).
        Default is ``True``.
    sum_method : str
        Method to sum the 95th percentile of the Hausdorff Distance. Default is ``max``, which argmax the 95th
        percentile of the Hausdorff Distance. If ``sum``, sum the 95th percentile of the Hausdorff Distance.

    Returns
    -------
    float
        Hausdorff Distance.

    Examples
    --------
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import hausdorff_distance_metric
    >>> import torch
    >>> datax = torch.randint(0, 2, (3, 2, 100, 100))
    >>> datay = torch.randint(0, 2, (3, 2, 100, 100))
    >>> hausdorff_distance_metric(datax, datay)
    5.858907404245753
    """
    if batched:
        hd = []
        for sl in range(x.shape[0]):
            hdx = x[sl].float().argmax(0).numpy() if sum_method == 'max' else x[sl].float().sum(0).numpy()
            hdy = y[sl].float().argmax(0).numpy() if sum_method == 'max' else y[sl].float().sum(0).numpy()
            hd.append(max(directed_hausdorff(hdx, hdy)[0], directed_hausdorff(hdy, hdx)[0]))
        return sum(hd) / len(hd)

    x = x.float().argmax(0).numpy() if sum_method == 'max' else x.float().sum(0).numpy()
    y = y.float().argmax(0).numpy() if sum_method == 'max' else y.float().sum(0).numpy()
    return max(directed_hausdorff(x, y)[0], directed_hausdorff(y, x)[0])


def hausdorff_distance_95_metric(x: torch.Tensor, y: torch.Tensor, batched: bool = True, sum_method='max') -> float:
    """Compute 95th percentile of the Hausdorff Distance.

    The Hausdorff distance is computed as the maximum between x to y and y to x.

    Parameters
    ----------
    x : torch.Tensor
        Ground Truth Tensor.
    y : torch.Tensor
        Prediction Tensor.
    batched : bool
        If True, compute Hausdorff distance for each batch and return a tensor with shape (batch_size,).
        If False, compute Hausdorff distance for the whole batch and return a tensor with shape (1,).
        Default is ``True``.
    sum_method : str
        Method to sum the 95th percentile of the Hausdorff Distance. Default is ``max``, which argmax the 95th
        percentile of the Hausdorff Distance. If ``sum``, sum the 95th percentile of the Hausdorff Distance.

    Returns
    -------
    float
        95th percentile of the Hausdorff Distance.

    Examples
    --------
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import hausdorff_distance_95_metric
    >>> import torch
    >>> datax = torch.randint(0, 2, (3, 2, 100, 100))
    >>> datay = torch.randint(0, 2, (3, 2, 100, 100))
    >>> hausdorff_distance_95_metric(datax, datay)
    5.853190166360368
    """
    if isinstance(x, np.ndarray):
        x = torch.from_numpy(x)
    if isinstance(y, np.ndarray):
        y = torch.from_numpy(y)

    if batched and (x.dim() == 4 and y.dim() == 4):
        hd = []
        for sl in range(x.shape[0]):
            hdx = x[sl].float().argmax(0).numpy() if sum_method == 'max' else x[sl].float().sum(0).numpy()
            hdy = y[sl].float().argmax(0).numpy() if sum_method == 'max' else y[sl].float().sum(0).numpy()
            hd1 = directed_hausdorff(hdx, hdy)[0]
            hd2 = directed_hausdorff(hdy, hdx)[0]
            hd.append(np.percentile(np.hstack((hd1, hd2)), 95))
        return sum(hd) / len(hd)

    x = x.float().argmax(0).numpy() if sum_method == 'max' else x.float().sum(0).numpy()
    y = y.float().argmax(0).numpy() if sum_method == 'max' else y.float().sum(0).numpy()
    hd1 = directed_hausdorff(x, y)[0]
    hd2 = directed_hausdorff(y, x)[0]
    return np.percentile(np.hstack((hd1, hd2)), 95)


def iou_metric(
    x: torch.Tensor,
    y: torch.Tensor,
    include_background: bool = True,
    ignore_empty: bool = True,
    reduction: Union[str, None] = "mean",
) -> float:
    """Compute Intersection over Union.

    Parameters
    ----------
    x : torch.Tensor
        Ground Truth Tensor.
    y : torch.Tensor
        Prediction Tensor.
    include_background : bool
        If True, include background in the computation. Default is ``True``.
    ignore_empty : bool
        If True, ignore empty slices. Default is ``True``.
    reduction : str or None
        If None, return a tensor with shape (batch_size,).
        If ``mean``, return the mean of the tensor.
        If ``sum``, return the sum of the tensor.
        Default is ``mean``.

    Returns
    -------
    float
        Intersection over Union.

    Examples
    --------
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import iou_metric
    >>> import torch
    >>> datax = torch.randint(0, 2, (3, 2, 100, 100))
    >>> datay = torch.randint(0, 2, (3, 2, 100, 100))
    >>> iou_metric(datax, datay)
    0.33478260040283203
    """
    if isinstance(x, np.ndarray):
        x = torch.from_numpy(x)
    if isinstance(y, np.ndarray):
        y = torch.from_numpy(y)

    if not include_background:
        if y.dim() == 1:
            warnings.warn("single channel prediction, `include_background=False` ignored.")
        else:
            # if skipping background, removing first channel
            x = x[:, 1:]
            y = y[:, 1:]

    if x.shape != y.shape:
        raise ValueError(f"Prediction and ground truth should have same shapes, got {y.shape} and {x.shape}.")

    # reducing only spatial dimensions (not batch nor channels)
    n_len = len(y.shape)
    reduce_axis = list(range(2, n_len))
    intersection = torch.sum(x * y, dim=reduce_axis)

    y_o = torch.sum(x, reduce_axis)
    y_y_o = torch.sum(y, dim=reduce_axis)
    union = y_o + y_y_o - intersection

    _max = 1.0 if not ignore_empty else float("nan")
    iou_score = torch.where(union > 0, (intersection) / union, torch.tensor(_max, device=y_o.device))
    iou_score, _ = do_metric_reduction(iou_score, reduction=reduction)  # type: ignore

    return iou_score.item()


def precision_metric(
    x: torch.Tensor,
    y: torch.Tensor,
    include_background: bool = True,
    average="none",
    mdmc_average="samplewise",
    reduction: Union[str, None] = "mean_batch",
) -> float:
    """Compute Precision Score.

    Parameters
    ----------
    x : torch.Tensor
        Ground Truth Tensor.
    y : torch.Tensor
        Prediction Tensor.
    include_background : bool
        If True, include background in the computation. Default is ``True``.
    average : str
        If ``none``, return a tensor with shape (batch_size,).
        If ``mean``, return the mean of the tensor.
        If ``sum``, return the sum of the tensor.
        Default is ``mean``.
    mdmc_average : str
        If ``samplewise``, return a tensor with shape (batch_size,).
        If ``global``, return the mean of the tensor.
        If ``none``, return the sum of the tensor.
        Default is ``samplewise``.
    reduction : str or None
        If None, return a tensor with shape (batch_size,).
        If ``mean``, return the mean of the tensor.
        If ``sum``, return the sum of the tensor.
        Default is ``mean_batch``.

    Returns
    -------
    float
        Precision Score.

    Examples
    --------
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import precision_metric
    >>> import torch
    >>> datax = torch.randint(0, 2, (3, 2, 100, 100))
    >>> datay = torch.randint(0, 2, (3, 2, 100, 100))
    >>> precision_metric(datax, datay)
    0.5005333423614502
    """
    if isinstance(x, np.ndarray):
        x = torch.from_numpy(x)
    if isinstance(y, np.ndarray):
        y = torch.from_numpy(y)

    if not include_background:
        if y.dim() == 1:
            warnings.warn("single channel prediction, `include_background=False` ignored.")
        else:
            # if skipping background, removing first channel
            x = x[:, 1:]
            y = y[:, 1:]

    x = x.type(torch.uint8)
    y = y.type(torch.uint8)

    if x.shape != y.shape:
        raise ValueError(f"Prediction and ground truth should have same shapes, got {y.shape} and {x.shape}.")

    # to one hot per class
    pr = []
    for i in range(y.shape[1]):
        precision_score = F.precision(
            one_hot(y[:, i].unsqueeze(1), num_classes=2),
            one_hot(x[:, i].unsqueeze(1), num_classes=2),
            task="binary",
            average=average,
            multidim_average=mdmc_average,
            num_classes=y.shape[1],
        )
        precision_score, _ = do_metric_reduction(precision_score, reduction=reduction)  # type: ignore
        pr.append(precision_score.item())
    return torch.mean(torch.tensor(pr)).item()


def recall_metric(
    x: torch.Tensor,
    y: torch.Tensor,
    include_background: bool = True,
    average="none",
    mdmc_average="samplewise",
    reduction: Union[str, None] = "mean_batch",
) -> float:
    """Compute Recall Score.

    Parameters
    ----------
    x : torch.Tensor
        Ground Truth Tensor.
    y : torch.Tensor
        Prediction Tensor.
    include_background : bool
        If True, include background in the computation. Default is ``True``.
    average : str
        If ``none``, return a tensor with shape (batch_size,).
        If ``mean``, return the mean of the tensor.
        If ``sum``, return the sum of the tensor.
        Default is ``mean``.
    mdmc_average : str
        If ``samplewise``, return a tensor with shape (batch_size,).
        If ``global``, return the mean of the tensor.
        If ``none``, return the sum of the tensor.
        Default is ``samplewise``.
    reduction : str or None
        If None, return a tensor with shape (batch_size,).
        If ``mean``, return the mean of the tensor.
        If ``sum``, return the sum of the tensor.
        Default is ``mean_batch``.

    Returns
    -------
    float
        Recall Score.

    Examples
    --------
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import recall_metric
    >>> import torch
    >>> datax = torch.randint(0, 2, (3, 2, 100, 100))
    >>> datay = torch.randint(0, 2, (3, 2, 100, 100))
    >>> recall_metric(datax, datay)
    0.5005333423614502
    """
    if isinstance(x, np.ndarray):
        x = torch.from_numpy(x)
    if isinstance(y, np.ndarray):
        y = torch.from_numpy(y)

    if not include_background:
        if y.dim() == 1:
            warnings.warn("single channel prediction, `include_background=False` ignored.")
        else:
            # if skipping background, removing first channel
            x = x[:, 1:]
            y = y[:, 1:]

    x = x.type(torch.uint8)
    y = y.type(torch.uint8)

    if x.shape != y.shape:
        raise ValueError(f"Prediction and ground truth should have same shapes, got {y.shape} and {x.shape}.")

    # to one hot per class
    rec = []
    for i in range(y.shape[1]):
        recall_score = F.recall(
            one_hot(y[:, i].unsqueeze(1), num_classes=2),
            one_hot(x[:, i].unsqueeze(1), num_classes=2),
            task="binary",
            average=average,
            multidim_average=mdmc_average,
            num_classes=y.shape[1],
        )
        recall_score, _ = do_metric_reduction(recall_score, reduction=reduction)  # type: ignore
        rec.append(recall_score.item())
    return torch.mean(torch.tensor(rec)).item()


def surface_distances(x, y, voxelspacing=None, connectivity=1):
    """The distances between the surface voxel of binary objects in result and their nearest partner surface voxel of a
    binary object in reference.

    Parameters
    ----------
    x : torch.Tensor
        Ground Truth Tensor.
    y : torch.Tensor
        Prediction Tensor.
    voxelspacing : Voxel Spacing. Defaults to ``None``.
    connectivity : int
        Connectivity. Defaults to ``1``.

    Returns
    -------
    np.ndarray
        The distances between the surface voxel of binary objects in result and their nearest partner surface voxel of
        a binary object in reference.

    Examples
    --------
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import surface_distances
    >>> import torch
    >>> datax = torch.randint(0, 2, (3, 2, 100, 100))
    >>> datay = torch.randint(0, 2, (3, 2, 100, 100))
    >>> surface_distances(datax, datay)
    array([0., 0., 1., ..., 0., 0., 0.])
    >>> surface_distances(datax, datay).mean()
    0.5083586894950562

    .. note::
        This function is based on the medpy implementation of the Average Symmetric Surface Distance (ASD) metric.
        Source: https://github.com/loli/medpy/blob/master/medpy/metric/binary.py#L458
    """
    x = np.atleast_1d(np.array([x]).astype(np.bool_))
    y = np.atleast_1d(np.array([y]).astype(np.bool_))

    if voxelspacing is not None:
        voxelspacing = _ni_support._normalize_sequence(voxelspacing, y.ndim)  # pylint: disable=protected-access
        voxelspacing = np.asarray(voxelspacing, dtype=np.float64)
        if not voxelspacing.flags.contiguous:
            voxelspacing = voxelspacing.copy()

    # binary structure
    footprint = generate_binary_structure(y.ndim, connectivity)

    # test for emptiness
    if np.count_nonzero(x) == 0:
        raise RuntimeError("The first supplied array does not contain any binary object.")
    if np.count_nonzero(y) == 0:
        raise RuntimeError("The second supplied array does not contain any binary object.")

    # extract only 1-pixel borderline of objects
    x_border = x ^ binary_erosion(x, structure=footprint, iterations=1)
    y_border = y ^ binary_erosion(y, structure=footprint, iterations=1)

    # compute average surface distance
    # Note: scipy distance transform is calculated only inside the borders of the foreground objects,
    # therefore the input has to be reversed
    dt = distance_transform_edt(~x_border, sampling=voxelspacing)
    sds = dt[y_border]

    return sds


class SegmentationMetrics:
    r"""Maintains running statistics for a given collection of segmentation metrics.

    Examples
    --------
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import SegmentationMetrics
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import \
    binary_cross_entropy_with_logits_metric
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import dice_metric
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import f1_per_class_metric
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import hausdorff_distance_metric
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import hausdorff_distance_95_metric
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import iou_metric
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import precision_metric
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import recall_metric
    >>> from atommic.collections.segmentation.metrics.segmentation_metrics import asd
    >>> import torch
    >>> datax = torch.randint(0, 2, (3, 2, 100, 100))
    >>> datay = torch.randint(0, 2, (3, 2, 100, 100))
    >>> metric_funcs = {
    ...     "binary_cross_entropy_with_logits": binary_cross_entropy_with_logits_metric,
    ...     "dice": dice_metric,
    ...     "f1_per_class": f1_per_class_metric,
    ...     "hausdorff_distance": hausdorff_distance_metric,
    ...     "hausdorff_distance_95": hausdorff_distance_95_metric,
    ...     "iou": iou_metric,
    ...     "precision": precision_metric,
    ...     "recall": recall_metric,
    ...     "asd": asd,
    ... }
    >>> metrics = SegmentationMetrics(metric_funcs)
    >>> metrics.push(datax, datay)
    >>> metrics.means()
    {'binary_cross_entropy_with_logits': 0.7527344822883606,
     'dice': 0.4993175268173218,
     'f1_per_class': 0.0,
     'hausdorff_distance': 5.8873012632302,
     'hausdorff_distance_95': 5.881561144720858,
     'iou': 0.3327365219593048,
     'precision': 0.5005833506584167,
     'recall': 0.5005833506584167,
     'asd': 0.503373912220483}
    >>> metrics.stddevs()
    {'binary_cross_entropy_with_logits': 0.0,
        'dice': 0.0,
        'f1_per_class': array([0., 0.]),
        'hausdorff_distance': 0.0,
        'hausdorff_distance_95': 0.0,
        'iou': 0.0,
        'precision': 0.0,
        'recall': 0.0,
        'asd': 0.0}
    >>> metrics.__repr__()
    'asd = 0.5034 +/- 0 binary_cross_entropy_with_logits = 0.7527 +/- 0 dice = 0.4993 +/- 0 f1_per_class = 0 +/- 0 \
    hausdorff_distance = 5.887 +/- 0 hausdorff_distance_95 = 5.882 +/- 0 iou = 0.3327 +/- 0 precision = 0.5006 +/- 0 \
    recall = 0.5006 +/- 0\n'
    """

    def __init__(self, metric_funcs):
        """Inits :class:`SegmentationMetrics`.
        Parameters
        ----------
        metric_funcs : dict
            A dict where the keys are metric names and the values are Python functions for evaluating that metric.
        """
        self.metric_funcs = metric_funcs
        self.metrics_scores = {metric: Statistics() for metric in metric_funcs}

    def push(self, x, y):
        """Pushes a new batch of metrics to the running statistics.

        Parameters
        ----------
        x : np.ndarray
            Target image. It must be a 3D array, where the first dimension is the number of slices. In case of 2D
            images, the first dimension should be 1.
        y : np.ndarray
            Predicted image. It must be a 3D array, where the first dimension is the number of slices. In case of 2D
            images, the first dimension should be 1.

        Returns
        -------
        dict
            A dict where the keys are metric names and the values are the computed metric scores.
        """
        for metric, func in self.metric_funcs.items():
            score = func(x, y)
            if isinstance(score, list):
                for i in enumerate(score):
                    if metric == f"F1_{i}":
                        self.metrics_scores[metric].push(score[i])
            else:
                self.metrics_scores[metric].push(score)

    def means(self):
        """Mean of the means of each metric."""
        return {metric: stat.mean() for metric, stat in self.metrics_scores.items()}

    def stddevs(self):
        """Standard deviation of the means of each metric."""
        return {metric: stat.stddev() for metric, stat in self.metrics_scores.items()}

    def __repr__(self):
        """Representation of the metrics."""
        means = self.means()
        stddevs = self.stddevs()
        metric_names = sorted(list(means))

        res = " ".join(f"{name} = {means[name]:.4g} +/- {2 * stddevs[name]:.4g}" for name in metric_names) + "\n"

        return res



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/timers.py

import time

import numpy as np
import torch

__all__ = ["NamedTimer"]


class NamedTimer:
    """A timer class that supports multiple named timers. A named timer can be used multiple times, in which case the
    average dt will be returned. A named timer cannot be started if it is already currently running.

    Use case: measuring execution of multiple code blocks.
    """

    _REDUCTION_TYPE = ["mean", "sum", "min", "max", "none"]

    def __init__(self, reduction="mean", sync_cuda=False, buffer_size=-1):
        """Inits :class:`NamedTimer`.

        Parameters
        ----------
        reduction : str
            Reduction over multiple timings of the same timer (none - returns the list instead of a scalar).
        sync_cuda : bool
            If True torch.cuda.synchronize() is called for start/stop.
        buffer_size : int
            If positive, limits the number of stored measures per name.
        """
        if reduction not in self._REDUCTION_TYPE:
            raise ValueError(f"Unknown reduction={reduction} please use one of {self._REDUCTION_TYPE}")

        self._reduction = reduction
        self._sync_cuda = sync_cuda
        self._buffer_size = buffer_size

        self.reset()

    def __getitem__(self, k):
        """Get item from timer."""
        return self.get(k)

    @property
    def buffer_size(self):
        """Returns the buffer size of the timer."""
        return self._buffer_size

    @property
    def _reduction_fn(self):
        """Returns the reduction function for the timer."""
        if self._reduction == "none":

            def fn(x):  # pragma: no cover
                return x

        else:
            fn = getattr(np, self._reduction)
        return fn

    def reset(self, name=None):
        """Resents all / specific timer

        Parameters
        ----------
        name : str
            Timer name to reset (if None all timers are reset).
        """
        if name is None:
            self.timers = {}
        else:
            self.timers[name] = {}

    def start(self, name=""):
        """Starts measuring a named timer.

        Parameters
        ----------
        name : str
            Timer name to start.
        """
        timer_data = self.timers.get(name, {})

        if "start" in timer_data:
            raise RuntimeError(f"Cannot start timer = '{name}' since it is already active")

        # synchronize pytorch cuda execution if supported
        if self._sync_cuda and torch.cuda.is_initialized():
            torch.cuda.synchronize()

        timer_data["start"] = time.time()

        self.timers[name] = timer_data

    def stop(self, name=""):
        """Stops measuring a named timer.

        Parameters
        ----------
        name : str
            Timer name to stop.
        """
        timer_data = self.timers.get(name)
        if (timer_data is None) or ("start" not in timer_data):
            raise RuntimeError(f"Cannot end timer = '{name}' since it is not active")

        # synchronize pytorch cuda execution if supported
        if self._sync_cuda and torch.cuda.is_initialized():
            torch.cuda.synchronize()

        # compute dt and make timer inactive
        dt = time.time() - timer_data.pop("start")

        # store dt
        timer_data["dt"] = timer_data.get("dt", []) + [dt]

        # enforce buffer_size if positive
        if self._buffer_size > 0:
            timer_data["dt"] = timer_data["dt"][-self._buffer_size :]

        self.timers[name] = timer_data

    def get(self, name=""):
        """Returns the value of a named timer

        Parameters
        ----------
        name : str
            Timer name to return.
        """
        dt_list = self.timers[name].get("dt", [])

        return self._reduction_fn(dt_list)

    def export(self):
        """Exports a dictionary with average/all dt per named timer"""
        fn = self._reduction_fn

        return {k: fn(v["dt"]) for k, v in self.timers.items() if "dt" in v}



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/optim/optimizer_with_master_params.py

from contextlib import contextmanager

import torch

from atommic.utils import logging

HAVE_APEX = False


def _zero_grad_group_helper(group, set_to_none):
    """Zero out the gradient for a group of parameters. Note: copied from torch.optim.optimizer."""
    for param in group:
        if param.grad is not None:
            if set_to_none:
                param.grad = None
            else:
                if param.grad.grad_fn is not None:
                    param.grad.detach_()
                else:
                    param.grad.requires_grad_(False)
                param.grad.zero_()


def _multi_tensor_copy_this_to_that(this, that, overflow_buf):
    """Use multi-tensor-applier to copy values from one list to another. We don't have a blfoat16 implementation so for
    now if the overflow_buf is not provided, we default back to simple loop copy to be compatible with bfloat16.
    """
    if overflow_buf:
        # Scaling with factor `1.0` is equivalent to copy.
        multi_tensor_applier(amp_C.multi_tensor_scale, overflow_buf, [this, that], 1.0)  # noqa: F821
    else:
        # FIXME: use multi-tensor applier for bf16
        for this_, that_ in zip(this, that):
            that_.copy_(this_)


class GradBucket:
    """Persistent buffer for main gradients that remains allocated between training iterations."""

    def __init__(self, numel, chunk_size_mb):
        """Inits :class:`GradBucket`.

        Parameters
        ----------
        numel : int
            Number of elements in the buffer.
        chunk_size_mb : int
            Chunk size in MB.
        """
        if not HAVE_APEX:
            raise ImportError("Apex was not found. Using model parallel models will error out.")

        self.numel = numel
        self.data = torch.zeros(self.numel, dtype=torch.float, device=torch.cuda.current_device(), requires_grad=False)

        self.chunk_size_mb = chunk_size_mb
        if self.chunk_size_mb > 0:
            chunk_size_bytes = chunk_size_mb * 1024 * 1024
            self.chunk_size_numel = chunk_size_bytes // 4
            self.num_chunks = self.numel // self.chunk_size_numel
            self.numel_per_chunk = [self.chunk_size_numel] * self.num_chunks
            if self.numel % self.chunk_size_numel != 0:
                self.num_chunks += 1
                self.numel_per_chunk.append(self.numel % self.chunk_size_numel)

            self.start_index_per_chunk = torch.cumsum(torch.tensor([0] + self.numel_per_chunk[:-1]), dim=0)
            self.current_chunk = 0
            self.computed_numel_per_chunk = [0] * self.num_chunks

    def zero(self):
        """Reset the buffer to zero."""
        self.data.zero_()

    def allreduce_buffer(self):
        """Synchronous buffer data allreduce"""
        self.data.div_(get_data_parallel_world_size())  # noqa: F821
        torch.distributed.all_reduce(self.data, group=get_data_parallel_group())  # noqa: F821

    def get(self, shape, start_index):
        """Return a tensor with the input `shape` as a view into the 1-D data starting at `start_index`."""
        end_index = start_index + shape.numel()
        if end_index > self.numel:
            raise AssertionError("requested tensor is out of the buffer range.")
        buffer_tensor = self.data[start_index:end_index]
        buffer_tensor = buffer_tensor.view(shape)

        grad_chunk_info = None
        if self.chunk_size_mb > 0:
            chunk = start_index // self.chunk_size_numel
            chunk_start_index = self.start_index_per_chunk[chunk]
            chunk_end_index = chunk_start_index + self.numel_per_chunk[chunk]
            grad_chunk_info = {chunk: min(chunk_end_index, end_index) - start_index}
            while chunk_end_index < end_index:
                chunk += 1
                chunk_start_index = self.start_index_per_chunk[chunk]
                chunk_end_index = chunk_start_index + self.numel_per_chunk[chunk]
                grad_chunk_info[chunk] = min(chunk_end_index, end_index) - chunk_start_index

        return buffer_tensor, grad_chunk_info

    def update_chunk_info(self, grad_chunk_info):
        """Update the chunk info with the grad_chunk_info."""
        for chunk in grad_chunk_info.keys():
            self.computed_numel_per_chunk[chunk] += grad_chunk_info[chunk]

    def get_allreduce_tensor(self):
        """Get a tensor that can be used for allreduce."""
        if self.computed_numel_per_chunk[self.current_chunk] != self.numel_per_chunk[self.current_chunk]:
            return None

        chunk_start_index = self.start_index_per_chunk[self.current_chunk]
        chunk_end_index = chunk_start_index + self.numel_per_chunk[self.current_chunk]
        self.computed_numel_per_chunk[self.current_chunk] = 0
        self.current_chunk += 1
        if self.current_chunk == self.num_chunks:
            self.current_chunk = 0

        return self.data[chunk_start_index:chunk_end_index]


class MainParamsOptimizerWrapper(torch.optim.Optimizer):
    """Float16 optimizer wrapper for half precision (fp16 and bf16) data types. This optimizer wrapper holds main
    parameters and gradients in fp32 to support stable convergence.
    """

    def __init__(  # noqa: MC0001
        self,
        optimizer,
        fp32_grad_accum=False,
        contiguous_grad_bucket=False,
        async_grad_allreduce=False,
        grad_div_ar_fusion=True,
        grad_allreduce_chunk_size_mb=0,
    ):
        """Inits :class:`MainParamsOptimizerWrapper`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Base optimizer such as Adam or SGD.
        fp32_grad_accum : bool
            To enable the use of fp32 in gradient accumulation and allreduce.
        contiguous_grad_bucket : bool
            To enable allocating the master gradients in the contiguous memory space to reduce memory fragmentation.
        async_grad_allreduce : bool
            Enable asynchronous gradient allreduce that is executed along with the training step back prop.
        """
        super().__init__(optimizer.param_groups)  # pylint: disable=no-value-for-parameter
        if not HAVE_APEX:
            raise ImportError("Apex was not found. Using model parallel models will error out.")

        self.optimizer = optimizer
        if not self.optimizer:
            raise AssertionError("no optimizer is provided.")
        if contiguous_grad_bucket and not fp32_grad_accum:
            raise AssertionError("contiguous gradient buffer assumes using fp32 grad.")
        if async_grad_allreduce:
            if not fp32_grad_accum:
                raise AssertionError(
                    "async allreduce applies to master gradients only, "
                    "which is supposed to be accumulated after grad op."
                )
            if not contiguous_grad_bucket:
                raise AssertionError("currently async_grad_allreduce is supported only with contiguous_grad_bucket.")

        self._fp32_grad_accum = fp32_grad_accum
        self._contiguous_grad_bucket = contiguous_grad_bucket
        self._async_grad_allreduce = async_grad_allreduce and get_data_parallel_world_size() > 1  # noqa: F821
        self._grad_divisor = 1 / get_data_parallel_world_size()  # noqa: F821

        if self._async_grad_allreduce:
            # use @no_sync to disable backward grad sync during gradient accumulation
            self._require_backward_grad_sync = True
            self._grad_div_ar_fusion = grad_div_ar_fusion
            self._grad_allreduce_chunk_size_mb = grad_allreduce_chunk_size_mb
        else:
            self._require_backward_grad_sync = False
            self._grad_div_ar_fusion = False
            self._grad_allreduce_chunk_size_mb = 0

        # Dummy tensor needed for apex multi-apply tensor.
        self._dummy_overflow_buf = None

        # Create persistent buffers for main gradients in contiguous memory space
        # - Chunked element-wise and allreduce ops without creating a temporary buffer for merged operation
        # - Low memory fragmentation
        self._main_grad_buffers = None
        if self._contiguous_grad_bucket:
            self._main_grad_buffers = {}
            # get the size of buffers
            num_elements = {}
            for i, param_group in enumerate(self.optimizer.param_groups):
                for param in param_group["params"]:
                    if param.requires_grad:
                        num_elements[i] = num_elements.get(i, 0) + param.data.nelement()

                # Allocate gradient memory buffers for each data type
                if any(param.requires_grad for param in param_group["params"]):
                    self._main_grad_buffers[i] = GradBucket(num_elements[i], self._grad_allreduce_chunk_size_mb)

        # Three groups of parameters:
        self.float16_groups = []  # original float16 parameters
        self.fp32_from_float16_groups = []  # fp32 copy of float16 parameters
        self.fp32_from_fp32_groups = []  # original fp32 parameters

        # gradient function hooks
        if self._fp32_grad_accum:
            self.grad_accs = []

        # For all the groups in the original optimizer:
        for i, param_group in enumerate(self.optimizer.param_groups):
            float16_params_this_group = []
            fp32_params_this_group = []
            fp32_from_float16_params_this_group = []
            # For all the parameters in this group:
            for j, param in enumerate(param_group["params"]):
                if param.requires_grad:
                    # float16 params:
                    if param.type() in ["torch.cuda.HalfTensor", "torch.cuda.BFloat16Tensor"]:
                        float16_params_this_group.append(param)

                        # Allocate the main parameter
                        main_param = param.detach().clone().float()

                        # Copy tensor model parallel attributes.
                        copy_tensor_model_parallel_attributes(main_param, param)  # noqa: F821
                        if hasattr(param, "shared"):
                            main_param.shared = param.shared

                        # Assign the grad buffer offset to main parameters
                        if self._contiguous_grad_bucket:
                            num_elements[i] = num_elements[i] - param.data.nelement()
                            main_param.grad, grad_chunk_info = self._main_grad_buffers[i].get(
                                param.data.shape, num_elements[i]
                            )
                            param.main_grad = main_param.grad

                        # Replace the optimizer params with the new fp32 copy.
                        param_group["params"][j] = main_param
                        fp32_from_float16_params_this_group.append(main_param)
                        # Reset existing state dict key to the new main param.
                        if param in self.optimizer.state:
                            self.optimizer.state[main_param] = self.optimizer.state.pop(param)
                    elif param.type() == "torch.cuda.FloatTensor":
                        fp32_params_this_group.append(param)
                        param_group["params"][j] = param

                    else:
                        raise TypeError(
                            "Wrapped parameters must be one of torch.cuda.FloatTensor,  torch.cuda.HalfTensor, "
                            f"or torch.cuda.BFloat16Tensor. Received {param.type()}"
                        )

                # Add gradient accumulation hook for fp32 grad accumulation
                if self._fp32_grad_accum and param.requires_grad:
                    # Expand so we get access to grad_fn.
                    param_tmp = param.expand_as(param)
                    # Get the gradient accumulator function.
                    grad_acc = param_tmp.grad_fn.next_functions[0][0]
                    grad_acc.register_hook(self._make_param_hook(param, main_param, i, grad_chunk_info))
                    self.grad_accs.append(grad_acc)

            self.float16_groups.append(float16_params_this_group)
            self.fp32_from_float16_groups.append(fp32_from_float16_params_this_group)
            self.fp32_from_fp32_groups.append(fp32_params_this_group)

        # init exp_avg and exp_avg_sq before loading optimizer state, needed for dist checkpointing
        self._init_opt_state()

        # Leverage state_dict() and load_state_dict() to
        # recast preexisting per-param state tensors
        self.optimizer.load_state_dict(self.optimizer.state_dict())

    def _make_param_hook(self, param, main_param, i, grad_chunk_info):
        """Create the grad accumulation and all-reduce hook for back prop."""

        def param_hook(*unused):  # pylint: disable=unused-argument
            """Gradient accumulation and all-reduce."""
            if param.grad is None:
                return
            if main_param.grad is None:
                main_param.grad = param.grad.float()
            else:
                main_param.grad.add_(param.grad.data)
            # Deallocate grad memory.
            param.grad = None

            # Asynchronous gradients allreduce across data_parallel ranks
            if self._require_backward_grad_sync:
                if self._grad_allreduce_chunk_size_mb > 0:
                    self._main_grad_buffers[i].update_chunk_info(grad_chunk_info)
                    while True:
                        allreduce_tensor = self._main_grad_buffers[i].get_allreduce_tensor()
                        if allreduce_tensor is None:
                            break
                        if self._grad_div_ar_fusion:
                            torch.distributed.all_reduce(
                                allreduce_tensor,
                                group=get_data_parallel_group(),  # noqa: F821
                                async_op=True,
                                op=torch.distributed._make_nccl_premul_sum(  # pylint: disable=protected-access
                                    self._grad_divisor
                                ),
                            )
                        else:
                            allreduce_tensor.div_(get_data_parallel_world_size())  # noqa: F821
                            torch.distributed.all_reduce(
                                allreduce_tensor,
                                group=get_data_parallel_group(),  # noqa: F821
                                async_op=True,
                            )
                else:
                    if self._grad_div_ar_fusion:
                        torch.distributed.all_reduce(
                            main_param.grad,
                            group=get_data_parallel_group(),  # noqa: F821
                            async_op=True,
                            op=torch.distributed._make_nccl_premul_sum(  # pylint: disable=protected-access
                                self._grad_divisor
                            ),
                        )
                    else:
                        main_param.grad.div_(get_data_parallel_world_size())  # noqa: F821
                        torch.distributed.all_reduce(
                            main_param.grad,
                            group=get_data_parallel_group(),  # noqa: F821
                            async_op=True,
                        )

        return param_hook

    def zero_grad(self, set_to_none=True):
        """We only need to zero the model related parameters, i.e., float16_groups & fp32_from_fp32_groups. We
        additionally zero fp32_from_float16_groups as a memory optimization to reduce fragmentation; in the case of
        set_to_none==True, the space used by this field can be safely deallocated at this point.
        """
        for group in self.float16_groups:
            _zero_grad_group_helper(group, set_to_none)
        if self._contiguous_grad_bucket:
            for i in self._main_grad_buffers:
                self._main_grad_buffers[i].zero()
        else:
            for group in self.fp32_from_float16_groups:
                _zero_grad_group_helper(group, set_to_none)
        for group in self.fp32_from_fp32_groups:
            _zero_grad_group_helper(group, set_to_none)

    def copy_model_grads_to_main_grads(self):
        """Copy model grads to main grads."""
        # This only needs to be done for the float16 group.
        for model_group, main_group in zip(self.float16_groups, self.fp32_from_float16_groups):
            for model_param, main_param in zip(model_group, main_group):
                if model_param.grad is not None:
                    main_param.grad = model_param.grad.float()

                # Safe to deallocate model's grad after copying.
                # (If using contiguous buffers, main_grad's memory should
                # persist and therefore should not be deallocated.)
                model_param.grad = None

    def _get_model_and_main_params_data_float16(self):
        """Get model and main params data in float16."""
        model_data = []
        main_data = []
        half_dtype = None
        for model_group, main_group in zip(self.float16_groups, self.fp32_from_float16_groups):
            for model_param, main_param in zip(model_group, main_group):
                if half_dtype is None:
                    half_dtype = model_param.data.dtype
                model_data.append(model_param.data)
                main_data.append(main_param.data)
        return model_data, main_data, half_dtype

    def _set_overflow_buffer(self, half_dtype):
        """Set overflow buffer."""
        if half_dtype == torch.float16:
            if self._dummy_overflow_buf is None:
                self._dummy_overflow_buf = torch.cuda.IntTensor([0])
            else:
                self._dummy_overflow_buf.fill_(0)

    def _copy_main_params_to_model_params(self):
        """Copy main params to model params."""
        # Only needed for the float16 params.
        model_data, main_data, half_dtype = self._get_model_and_main_params_data_float16()
        self._set_overflow_buffer(half_dtype)
        _multi_tensor_copy_this_to_that(this=main_data, that=model_data, overflow_buf=self._dummy_overflow_buf)

    def _copy_model_params_to_main_params(self):
        """Copy model params to main params."""
        # Only needed for the float16 params.
        model_data, main_data, half_dtype = self._get_model_and_main_params_data_float16()
        self._set_overflow_buffer(half_dtype)
        _multi_tensor_copy_this_to_that(this=model_data, that=main_data, overflow_buf=self._dummy_overflow_buf)

    def reload_model_params(self):
        """Reload model params."""
        self._copy_model_params_to_main_params()

    # pylint: disable=arguments-differ
    @torch.no_grad()
    def step(self, **kwargs):
        """Step the optimizer."""
        # While async grad allreduce is enabled, bprop will keep moving forward without waiting for the finish of
        # async grad AR works. Hence, to guarantee the correctness of grads reduction, we cannot start weight update
        # until all async grad AR works are done.
        if self._async_grad_allreduce:
            torch.cuda.synchronize()
        self.optimizer.step(closure=None, **kwargs)
        # Update params from main params.
        with torch.no_grad():
            self._copy_main_params_to_model_params()
        # Successful update.
        return True

    def state_dict(self):
        """Return the state of the optimizer."""
        return {"optimizer": self.optimizer.state_dict(), "fp32_from_fp16_params": self.fp32_from_float16_groups}

    def load_state_dict(self, state_dict):
        """Load the state of the optimizer."""
        # Optimizer.
        optimizer_key = "optimizer"
        if optimizer_key not in state_dict:
            optimizer_key = "optimizer_state_dict"
            logging.info("***WARNING*** loading optimizer from an old checkpoint ...")
        self.optimizer.load_state_dict(state_dict[optimizer_key])

        # Copy data for the main params.
        fp32_from_float16_params_key = "fp32_from_fp16_params"
        if fp32_from_float16_params_key not in state_dict:
            fp32_from_float16_params_key = "fp32_from_fp16"
        for current_group, saved_group in zip(self.fp32_from_float16_groups, state_dict[fp32_from_float16_params_key]):
            for current_param, saved_param in zip(current_group, saved_group):
                current_param.data.copy_(saved_param.data)

    def allreduce_main_grads(self):
        """All reduce main grads."""
        for i in self._main_grad_buffers:
            self._main_grad_buffers[i].allreduce_buffer()

    @contextmanager
    def no_sync(self):
        """A context manager to disable gradient synchronizations across data-parallel ranks."""
        old_require_backward_grad_sync = self._require_backward_grad_sync
        self._require_backward_grad_sync = False
        try:
            yield
        finally:
            self._require_backward_grad_sync = old_require_backward_grad_sync

    @property
    def async_master_grads_allreduce(self):
        """Return whether to use async allreduce for master grads."""
        return self._async_grad_allreduce

    @property
    def fp32_grad_accumulation(self):
        """Return whether to accumulate gradients in fp32."""
        return self._fp32_grad_accum

    def get_parameters(self):
        """Return the parameters of the optimizer."""
        params = []
        for param_group in self.optimizer.param_groups:
            for param in param_group["params"]:
                if param.grad is not None:  # added to enable pp>1 training for adapters
                    params.append(param)
        return params

    def _get_state(self):
        """Promote state, so it can be retrieved or set via "optimizer_instance.state."""
        return self.optimizer.state if hasattr(self, "optimizer") else []

    def _set_state(self, value):
        """Promote state, so it can be retrieved or set via "optimizer_instance.state."""
        self.optimizer.state = value

    state = property(_get_state, _set_state)

    def _get_param_groups(self):
        """Promote param_groups, so it can be retrieved or set via "optimizer_instance.param_groups. For example, to
        adjust the learning rate.
        """
        return self.optimizer.param_groups if hasattr(self, "optimizer") else []

    def _set_param_groups(self, value):
        """Set param_groups."""
        self.optimizer.param_groups = value

    param_groups = property(_get_param_groups, _set_param_groups)

    def _get_defaults(self):
        """Promote defaults, so it can be retrieved or set via 'optimizer_instance.default'."""
        return self.optimizer.defaults if hasattr(self, "optimizer") else []

    def _set_defaults(self, value):
        """Set defaults."""
        self.optimizer.defaults = value

    defaults = property(_get_defaults, _set_defaults)
